# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main.ui',
# licensing of 'main.ui' applies.
#
# Created: Fri Mar 27 12:26:29 2020
#      by: pyside2-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_MondoMainWindow(object):
    def setupUi(self, MondoMainWindow):
        MondoMainWindow.setObjectName("MondoMainWindow")
        MondoMainWindow.resize(617, 351)
        self.centralwidget = QtWidgets.QWidget(MondoMainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.groupBox = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.psdButton = QtWidgets.QPushButton(self.groupBox)
        self.psdButton.setMinimumSize(QtCore.QSize(175, 0))
        self.psdButton.setObjectName("psdButton")
        self.verticalLayout.addWidget(self.psdButton)
        self.singleChannelPsdButton = QtWidgets.QPushButton(self.groupBox)
        self.singleChannelPsdButton.setObjectName("singleChannelPsdButton")
        self.verticalLayout.addWidget(self.singleChannelPsdButton)
        self.singleSlicePsdButton = QtWidgets.QPushButton(self.groupBox)
        self.singleSlicePsdButton.setObjectName("singleSlicePsdButton")
        self.verticalLayout.addWidget(self.singleSlicePsdButton)
        self.overlaidPsdButton = QtWidgets.QPushButton(self.groupBox)
        self.overlaidPsdButton.setMinimumSize(QtCore.QSize(175, 0))
        self.overlaidPsdButton.setObjectName("overlaidPsdButton")
        self.verticalLayout.addWidget(self.overlaidPsdButton)
        spacerItem = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.gridLayout.addWidget(self.groupBox, 0, 2, 1, 1)
        self.groupBox_4 = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox_4.setObjectName("groupBox_4")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox_4)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.csvButton = QtWidgets.QPushButton(self.groupBox_4)
        self.csvButton.setMinimumSize(QtCore.QSize(175, 0))
        self.csvButton.setObjectName("csvButton")
        self.verticalLayout_4.addWidget(self.csvButton)
        self.csvDownsampledButton = QtWidgets.QPushButton(self.groupBox_4)
        self.csvDownsampledButton.setMinimumSize(QtCore.QSize(175, 0))
        self.csvDownsampledButton.setObjectName("csvDownsampledButton")
        self.verticalLayout_4.addWidget(self.csvDownsampledButton)
        spacerItem1 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_4.addItem(spacerItem1)
        self.gridLayout.addWidget(self.groupBox_4, 1, 0, 1, 1)
        self.groupBox_2 = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.spectrogramButton = QtWidgets.QPushButton(self.groupBox_2)
        self.spectrogramButton.setMinimumSize(QtCore.QSize(175, 0))
        self.spectrogramButton.setObjectName("spectrogramButton")
        self.verticalLayout_2.addWidget(self.spectrogramButton)
        spacerItem2 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem2)
        self.gridLayout.addWidget(self.groupBox_2, 0, 1, 1, 1)
        self.groupBox_3 = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox_3.setObjectName("groupBox_3")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_3)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.timeButton = QtWidgets.QPushButton(self.groupBox_3)
        self.timeButton.setMinimumSize(QtCore.QSize(175, 0))
        self.timeButton.setObjectName("timeButton")
        self.verticalLayout_3.addWidget(self.timeButton)
        self.synchronousTimeButton = QtWidgets.QPushButton(self.groupBox_3)
        self.synchronousTimeButton.setMinimumSize(QtCore.QSize(175, 0))
        self.synchronousTimeButton.setObjectName("synchronousTimeButton")
        self.verticalLayout_3.addWidget(self.synchronousTimeButton)
        self.overlaidTimeButton = QtWidgets.QPushButton(self.groupBox_3)
        self.overlaidTimeButton.setMinimumSize(QtCore.QSize(175, 0))
        self.overlaidTimeButton.setObjectName("overlaidTimeButton")
        self.verticalLayout_3.addWidget(self.overlaidTimeButton)
        spacerItem3 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem3)
        self.gridLayout.addWidget(self.groupBox_3, 0, 0, 1, 1)
        self.groupBox_5 = QtWidgets.QGroupBox(self.centralwidget)
        self.groupBox_5.setObjectName("groupBox_5")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.groupBox_5)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.fileInformationButton = QtWidgets.QPushButton(self.groupBox_5)
        self.fileInformationButton.setObjectName("fileInformationButton")
        self.verticalLayout_6.addWidget(self.fileInformationButton)
        self.viewSettingsButton = QtWidgets.QPushButton(self.groupBox_5)
        self.viewSettingsButton.setObjectName("viewSettingsButton")
        self.verticalLayout_6.addWidget(self.viewSettingsButton)
        self.rawExportButton = QtWidgets.QPushButton(self.groupBox_5)
        self.rawExportButton.setObjectName("rawExportButton")
        self.verticalLayout_6.addWidget(self.rawExportButton)
        self.splitFileButton = QtWidgets.QPushButton(self.groupBox_5)
        self.splitFileButton.setObjectName("splitFileButton")
        self.verticalLayout_6.addWidget(self.splitFileButton)
        spacerItem4 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_6.addItem(spacerItem4)
        self.gridLayout.addWidget(self.groupBox_5, 1, 1, 1, 1)
        self.verticalLayout_5.addLayout(self.gridLayout)
        spacerItem5 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_5.addItem(spacerItem5)
        self.verticalLayout_5.setStretch(1, 1)
        MondoMainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MondoMainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 617, 21))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuHelp = QtWidgets.QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuCheckForUpdates = QtWidgets.QMenu(self.menuHelp)
        self.menuCheckForUpdates.setObjectName("menuCheckForUpdates")
        self.menuSettings = QtWidgets.QMenu(self.menubar)
        self.menuSettings.setObjectName("menuSettings")
        MondoMainWindow.setMenuBar(self.menubar)
        self.actionAbout = QtWidgets.QAction(MondoMainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionExit = QtWidgets.QAction(MondoMainWindow)
        self.actionExit.setObjectName("actionExit")
        self.actionAboutLibraries = QtWidgets.QAction(MondoMainWindow)
        self.actionAboutLibraries.setObjectName("actionAboutLibraries")
        self.actionUpdateLatestStable = QtWidgets.QAction(MondoMainWindow)
        self.actionUpdateLatestStable.setObjectName("actionUpdateLatestStable")
        self.actionUpdateSpecificBranch = QtWidgets.QAction(MondoMainWindow)
        self.actionUpdateSpecificBranch.setObjectName("actionUpdateSpecificBranch")
        self.actionUpdateSpecificCommit = QtWidgets.QAction(MondoMainWindow)
        self.actionUpdateSpecificCommit.setObjectName("actionUpdateSpecificCommit")
        self.actionUpdateCurrentBranch = QtWidgets.QAction(MondoMainWindow)
        self.actionUpdateCurrentBranch.setObjectName("actionUpdateCurrentBranch")
        self.actionPreferences = QtWidgets.QAction(MondoMainWindow)
        self.actionPreferences.setObjectName("actionPreferences")
        self.menuFile.addAction(self.actionExit)
        self.menuCheckForUpdates.addAction(self.actionUpdateLatestStable)
        self.menuCheckForUpdates.addAction(self.actionUpdateCurrentBranch)
        self.menuCheckForUpdates.addAction(self.actionUpdateSpecificBranch)
        self.menuCheckForUpdates.addAction(self.actionUpdateSpecificCommit)
        self.menuHelp.addAction(self.menuCheckForUpdates.menuAction())
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionAbout)
        self.menuHelp.addAction(self.actionAboutLibraries)
        self.menuSettings.addAction(self.actionPreferences)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuSettings.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MondoMainWindow)
        QtCore.QObject.connect(self.actionExit, QtCore.SIGNAL("triggered()"), MondoMainWindow.close)
        QtCore.QMetaObject.connectSlotsByName(MondoMainWindow)

    def retranslateUi(self, MondoMainWindow):
        MondoMainWindow.setWindowTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Mondo", None, -1))
        self.groupBox.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Frequency Analysis", None, -1))
        self.psdButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Power Spectral Density (PSD)", None, -1))
        self.singleChannelPsdButton.setToolTip(QtWidgets.QApplication.translate("MondoMainWindow", "Overlaid PSD streamlined for plotting a single channel across multiple time slices", None, -1))
        self.singleChannelPsdButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Single Channel Overlaid PSD", None, -1))
        self.singleSlicePsdButton.setToolTip(QtWidgets.QApplication.translate("MondoMainWindow", "Overlaid PSD streamlined for plotting multiple channels in a single time slice", None, -1))
        self.singleSlicePsdButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Single Time Slice Overlaid PSD", None, -1))
        self.overlaidPsdButton.setToolTip(QtWidgets.QApplication.translate("MondoMainWindow", "Ovelaid PSD for plotting arbitrary channels and time slices ", None, -1))
        self.overlaidPsdButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Overlaid PSD", None, -1))
        self.groupBox_4.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Raw Data", None, -1))
        self.csvButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "CSV Output", None, -1))
        self.csvDownsampledButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "CSV Downsampled Output", None, -1))
        self.groupBox_2.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Time/Frequency Analysis", None, -1))
        self.spectrogramButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Spectrogram", None, -1))
        self.groupBox_3.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Time Analysis", None, -1))
        self.timeButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Time Data", None, -1))
        self.synchronousTimeButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Synchronous Time Data", None, -1))
        self.overlaidTimeButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Overlaid Time Data", None, -1))
        self.groupBox_5.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "File Manipulation", None, -1))
        self.fileInformationButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "File Information", None, -1))
        self.viewSettingsButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "View Settings", None, -1))
        self.rawExportButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Advanced Raw Export", None, -1))
        self.splitFileButton.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Rend File Asunder", None, -1))
        self.menuFile.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "File", None, -1))
        self.menuHelp.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Help", None, -1))
        self.menuCheckForUpdates.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Check for Updates", None, -1))
        self.menuSettings.setTitle(QtWidgets.QApplication.translate("MondoMainWindow", "Settings", None, -1))
        self.actionAbout.setText(QtWidgets.QApplication.translate("MondoMainWindow", "About Mondo...", None, -1))
        self.actionAbout.setToolTip(QtWidgets.QApplication.translate("MondoMainWindow", "About Mondo...", None, -1))
        self.actionExit.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Exit", None, -1))
        self.actionAboutLibraries.setText(QtWidgets.QApplication.translate("MondoMainWindow", "About Libraries...", None, -1))
        self.actionAboutLibraries.setToolTip(QtWidgets.QApplication.translate("MondoMainWindow", "About Libraries...", None, -1))
        self.actionUpdateLatestStable.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Latest Stable", None, -1))
        self.actionUpdateSpecificBranch.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Specific Branch...", None, -1))
        self.actionUpdateSpecificCommit.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Specific Commit..", None, -1))
        self.actionUpdateCurrentBranch.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Latest branch", None, -1))
        self.actionPreferences.setText(QtWidgets.QApplication.translate("MondoMainWindow", "Preferences...", None, -1))
        self.actionPreferences.setToolTip(QtWidgets.QApplication.translate("MondoMainWindow", "Preferences...", None, -1))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'preferences.ui',
# licensing of 'preferences.ui' applies.
#
# Created: Fri Mar 27 12:26:29 2020
#      by: pyside2-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_PreferencesDialog(object):
    def setupUi(self, PreferencesDialog):
        PreferencesDialog.setObjectName("PreferencesDialog")
        PreferencesDialog.resize(414, 379)
        self.verticalLayout = QtWidgets.QVBoxLayout(PreferencesDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(PreferencesDialog)
        self.tabWidget.setObjectName("tabWidget")
        self.unitTab = QtWidgets.QWidget()
        self.unitTab.setObjectName("unitTab")
        self.gridLayout = QtWidgets.QGridLayout(self.unitTab)
        self.gridLayout.setObjectName("gridLayout")
        self.unitsLabel = QtWidgets.QLabel(self.unitTab)
        font = QtGui.QFont()
        font.setWeight(75)
        font.setBold(True)
        self.unitsLabel.setFont(font)
        self.unitsLabel.setObjectName("unitsLabel")
        self.gridLayout.addWidget(self.unitsLabel, 0, 0, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(20, 13, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 1, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 4, 0, 1, 2)
        self.unitPreferences = UnitPreferencesWidget(self.unitTab)
        self.unitPreferences.setObjectName("unitPreferences")
        self.gridLayout.addWidget(self.unitPreferences, 1, 1, 1, 1)
        self.tabWidget.addTab(self.unitTab, "")
        self.verticalLayout.addWidget(self.tabWidget)
        self.buttonBox = QtWidgets.QDialogButtonBox(PreferencesDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(PreferencesDialog)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), PreferencesDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), PreferencesDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(PreferencesDialog)
        PreferencesDialog.setTabOrder(self.tabWidget, self.buttonBox)

    def retranslateUi(self, PreferencesDialog):
        PreferencesDialog.setWindowTitle(QtWidgets.QApplication.translate("PreferencesDialog", "Preferences", None, -1))
        self.unitsLabel.setText(QtWidgets.QApplication.translate("PreferencesDialog", "Display Units", None, -1))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.unitTab), QtWidgets.QApplication.translate("PreferencesDialog", "Units", None, -1))

from hyperborea.unit_preferences import UnitPreferencesWidget



import io
import os.path
import subprocess

from pyside2uic import compileUi


def create_py_from_ui(ui_file):
    output = io.StringIO()
    head, tail = os.path.split(os.path.abspath(ui_file))
    cwd = os.getcwd()
    try:
        os.chdir(head)
        compileUi(tail, output)
    finally:
        os.chdir(cwd)
    return output.getvalue()


def compare_and_update(filename, new_contents):
    

    try:
        with open(filename, 'r') as f:
            old_lines = f.readlines()
        old_uncommented = [l for l in old_lines if not l.startswith("#")]
        
        new_lines = new_contents.splitlines(keepends=True)
        new_uncommented = [l for l in new_lines if not l.startswith("#")]
        
        no_change = (new_uncommented == old_uncommented)
    except FileNotFoundError:
        no_change = False

    if no_change:
        print("No Change: {}".format(os.path.basename(filename)))
        return
    else:
        with open(filename, 'w') as f:
            f.write(new_contents)
        print("Updated: {}".format(os.path.basename(filename)))


def update_ui(ui_file):
    # pull off the .ui
    root, _ext = os.path.splitext(ui_file)
    if _ext != ".ui":
        root = ui_file
    
    # add ui_ to the front and .py to the end
    head, tail = os.path.split(root)
    output_file = os.path.join(head, "ui_" + tail + ".py")

    new_contents = create_py_from_ui(ui_file)
    compare_and_update(output_file, new_contents)


def main():
    for root, dirs, files in os.walk("."):
        # skip hidden files and folders
        files = [f for f in files if not f[0] == '.']
        dirs[:] = [d for d in dirs if not d[0] == '.']

        # skip build directory
        if 'build' in dirs:
            dirs.remove('build')

        for file in files:
            if file.endswith(".ui"):
                update_ui(os.path.join(root, file))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging.handlers
import multiprocessing
import os
import sys
import time

from PySide2 import QtCore, QtGui, QtWidgets

import mondo

# wrapper around "import matplotlib" that captures any commands for later use
from mondo.export_script import matplotlib_wrapper as matplotlib

from .main import MondoMainWindow
from . import mondo_rc  # needed for the icon

logger = logging.getLogger(__name__)


def main_is_frozen():
    """Return True if the script is frozen, False otherwise."""
    return getattr(sys, 'frozen', False)


def get_main_dir():
    """Return the path of the main script's directory, even when frozen."""
    if main_is_frozen():
        return os.path.dirname(sys.executable)
    return os.path.dirname(os.path.abspath(sys.argv[0]))


def setup_logging():
    """Configure logging for the whole program."""

    def my_excepthook(exctype, value, traceback):
        """Log the caught exception using the logging module."""
        exc_info = (exctype, value, traceback)
        logger.error("Uncaught Exception", exc_info=exc_info)

    sys.excepthook = my_excepthook

    # get an appropriate location for the log file
    logdir = QtCore.QStandardPaths.writableLocation(
        QtCore.QStandardPaths.DataLocation)
    logfile = os.path.join(logdir, "main.log")

    # make sure the directory exists
    os.makedirs(logdir, exist_ok=True)

    # create a log file handler
    file_log_handler = logging.handlers.RotatingFileHandler(logfile,
                                                            maxBytes=10e6,
                                                            backupCount=9)
    file_formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_formatter.default_time_format = '%Y-%m-%dT%H:%M:%S'
    file_formatter.default_msec_format = '%s,%03dZ'
    file_formatter.converter = time.gmtime
    file_log_handler.setFormatter(file_formatter)
    file_log_handler.setLevel(logging.DEBUG)

    root_logger = logging.getLogger()
    root_logger.addHandler(file_log_handler)
    root_logger.setLevel(logging.DEBUG)

    # remove pyusb's logging info
    pyusb_logger = logging.getLogger("usb")
    pyusb_logger.propagate = False

    # remove matplotlib debug level
    matplotlib_logger = logging.getLogger("matplotlib")
    matplotlib_logger.setLevel(logging.INFO)


def main():
    """Run the program."""

    # freeze_support() MUST be first. Anything before this will cease to exist
    multiprocessing.freeze_support()

    # work around for requests library to find root certificates
    if main_is_frozen():
        cacert = os.path.join(os.path.dirname(sys.executable), 'cacert.pem')
        os.environ["REQUESTS_CA_BUNDLE"] = cacert

    if not sys.stdout or not sys.stderr:
        sys.stdout = open(os.devnull)
        sys.stderr = open(os.devnull)

    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName("Mondo")
    app.setOrganizationDomain("suprocktech.com")
    app.setOrganizationName("Suprock Tech")

    # load the application icon
    icon = QtGui.QIcon()
    icon_reader = QtGui.QImageReader(":/mondo.ico")
    while True:
        pixmap = QtGui.QPixmap.fromImage(icon_reader.read())
        icon.addPixmap(pixmap)
        if not icon_reader.jumpToNextImage():
            break
    app.setWindowIcon(icon)

    # force the settings to use an INI file instead of the registry
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)

    # remove "What's This" button across the whole application
    app.setAttribute(QtCore.Qt.AA_DisableWindowContextHelpButton)

    app.setApplicationVersion(mondo.__version__)

    setup_logging()

    logger.info("Mondo started (Version {})".format(mondo.__version__))

    # fix for matplotlib unicode characters like Omega
    matplotlib.rc('font', size=17)

    mainwin = MondoMainWindow()
    mainwin.show()
    app.exec_()

    logger.info("Mondo finished")


if __name__ == '__main__':
    main()



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging

from PySide2 import QtWidgets

from . import util
from .. import export_script

logger = logging.getLogger(__name__)


def time_analysis(parent=None):
    ret = util.load_batch(parent=parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    channel_index = util.choose_channel(header, parent=parent)
    if channel_index is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, [channel_index],
                                   parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, parent=parent)
    if not ret:
        return  # user cancelled

    sequence_info, unit_info = util.sequence_data(batch_info, chunk=False,
                                                  parent=parent)

    time, data, _start, _end = sequence_info[channel_index][0]

    show_legend = False

    if data.shape[1] != 1:
        subchannel_index = util.choose_subchannel(batch_info[channel_index],
                                                  allow_all=True)
        if subchannel_index is None:
            return  # user cancelled
        elif subchannel_index != -1:
            data = data[:, subchannel_index]
        else:
            show_legend = True

        # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    ax.plot(time, data)

    ax.set_xlabel("Time (s)")
    unit_str = unit_info[channel_index]['utf8']
    if unit_str:
        ax.set_ylabel("Magnitude ({})".format(unit_str))
    else:
        ax.set_ylabel("Magnitude")

    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    ax.get_yaxis().get_major_formatter().set_useOffset(False)

    if show_legend:
        channel_data = batch_info[channel_index]
        ax.legend(channel_data.channel_decoder.subchannel_names)

    fig.show()


def synchronous_time_analysis(parent=None):
    ret = util.load_batch(parent=parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    channel_indexes = util.choose_synchronous_channels(header, parent)
    if channel_indexes is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, channel_indexes,
                                   parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, once=True, parent=parent)
    if not ret:
        return  # user cancelled

    sequence_info, unit_info = util.sequence_data(batch_info, chunk=False,
                                                  parent=parent)

    # group channels by unit type
    plot_groups = []
    groups_by_unit_type = {}
    for channel_index in sorted(channel_indexes):
        channel = header['channels'][channel_index]
        if channel.unit_type == 0:
            # don't share plots between channels with unknown units
            plot_groups.append([channel_index])
        elif channel.unit_type in groups_by_unit_type:
            # already have a group made
            plot_group = groups_by_unit_type[channel.unit_type]
            plot_group.append(channel_index)
        else:
            # need a new group
            plot_group = [channel_index]
            plot_groups.append(plot_group)
            groups_by_unit_type[channel.unit_type] = plot_group

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax_list = export_script.subplots_wrapper(len(plot_groups),
                                                  squeeze=False, sharex=True)

    for i, ax in enumerate(ax_list[:, 0]):
        plot_group = plot_groups[i]

        legend_strs = []

        for channel_index in plot_group:
            time, data, _start, _end = sequence_info[channel_index][0]
            ax.plot(time, data)

            channel_data = batch_info[channel_index]
            legend_strs.extend(channel_data.channel_decoder.subchannel_names)

        ax.set_xlabel("Time (s)")
        unit_str = unit_info[plot_group[0]]['utf8']
        if unit_str:
            ax.set_ylabel("Magnitude ({})".format(unit_str))
        else:
            ax.set_ylabel("Magnitude")

        ax.get_xaxis().get_major_formatter().set_useOffset(False)
        ax.get_yaxis().get_major_formatter().set_useOffset(False)

        ax.legend(legend_strs)

    fig.show()


def overlaid_time_analysis(parent=None):
    sequences = []
    subchannel_indexes = []
    unit_types = []
    unit_names = []
    names = []

    while True:
        # load a batch of files
        ret = util.load_batch(parent=parent)
        if ret is None:
            return  # error message displayed, or user cancelled
        file_infos, header = ret

        # choose a channel
        channel_index = util.choose_channel(header, parent=parent)
        if channel_index is None:
            return  # error message displayed, or user cancelled

        batch_info = util.decode_batch(file_infos, header, [channel_index],
                                       parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        subchannel_index = util.choose_subchannel(
            batch_info[channel_index], allow_all=False)
        if subchannel_index is None:
            return  # user cancelled
        subchannel_indexes.append(subchannel_index)

        batch_info = util.get_datetime_subset(batch_info, parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        ret = util.warn_about_lost_packets(batch_info, parent=parent)
        if not ret:
            return  # user cancelled

        sequence_info, unit_info = util.sequence_data(batch_info, chunk=False,
                                                      parent=parent)

        sequence = sequence_info[channel_index][0]

        sequences.append(sequence)
        unit_types.append(header['channels'][channel_index].unit_type)
        unit_names.append(unit_info[channel_index]['utf8'])

        # ask the user for a name for this set
        msg = "Display name for this batch (#{}):".format(len(sequences))
        name, ok = QtWidgets.QInputDialog.getText(parent, "Batch Name", msg)
        if not ok:
            return  # user cancelled

        names.append(name.strip())

        # ask the user if they want to load more
        msg = "Load more batches? Loaded {} so far".format(len(sequences))
        ret = QtWidgets.QMessageBox.question(
            parent, "More Batches?", msg,
            buttons=(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No),
            default=QtWidgets.QMessageBox.No)
        if ret != QtWidgets.QMessageBox.Yes:
            break

    # group channels by unit type
    plot_groups = []
    groups_by_unit_type = {}
    for batch_index, unit_type in enumerate(unit_types):
        if unit_type == 0:
            # don't share plots between channels with unknown units
            plot_groups.append([batch_index])
        elif unit_type in groups_by_unit_type:
            # already have a group made
            plot_group = groups_by_unit_type[unit_type]
            plot_group.append(batch_index)
        else:
            # need a new group
            plot_group = [batch_index]
            plot_groups.append(plot_group)
            groups_by_unit_type[unit_type] = plot_group

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax_list = export_script.subplots_wrapper(len(plot_groups),
                                                  squeeze=False, sharex=True)

    for i, ax in enumerate(ax_list[:, 0]):
        plot_group = plot_groups[i]

        for batch_index in plot_group:
            time, data, _start, _end = sequences[batch_index]
            subchannel_index = subchannel_indexes[batch_index]
            ax.plot(time, data[:, subchannel_index], label=names[batch_index])

        ax.legend()

        ax.set_xlabel("Time (s)")
        unit_str = unit_names[plot_group[0]]
        if unit_str:
            ax.set_ylabel("Magnitude ({})".format(unit_str))
        else:
            ax.set_ylabel("Magnitude")

        ax.get_xaxis().get_major_formatter().set_useOffset(False)
        ax.get_yaxis().get_major_formatter().set_useOffset(False)

    fig.show()



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import sys

from PySide2 import QtCore, QtWidgets

from . import __version__ as version
from .ui_about import Ui_AboutDialog

logger = logging.getLogger(__name__)


class AboutDialog(QtWidgets.QDialog, Ui_AboutDialog):
    def __init__(self, parent=None):
        super().__init__(parent, QtCore.Qt.MSWindowsFixedSizeDialogHint)

        self.setupUi(self)

        is_frozen = getattr(sys, 'frozen', False)
        if is_frozen:
            # load the build_info.txt
            main_dir = os.path.dirname(sys.executable)
            build_info_filename = os.path.join(main_dir, "build_info.txt")
            try:
                with open(build_info_filename, "r") as f:
                    lines = f.readlines()
                    build_date = lines[3].strip()
            except:
                logger.exception('Could not read build_info.txt')
                build_date = "ERROR"
        else:
            build_date = "Unknown"

        label_str = self.softwareLabel.text().format(version, build_date)
        self.softwareLabel.setText(label_str)

        self.layout().setSizeConstraint(QtWidgets.QLayout.SetFixedSize)



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging

from PySide2 import QtWidgets

from .ui_select_synchronous import Ui_SelectSynchronousDialog

logger = logging.getLogger(__name__)


class SelectSynchronousDialog(QtWidgets.QDialog, Ui_SelectSynchronousDialog):
    def __init__(self, streams, channels, parent=None):
        super().__init__(parent)

        self.streams = streams
        self.channels = channels

        # these are ordered by display order
        self.stream_buttons = []  # radio buttons
        self.channel_groups = []  # {ch_index: check box}

        self.widgets = []  # to prevent premature garbage collection

        self.setupUi(self)

        self.add_buttons()

    def add_buttons(self):
        d = {}  # key: min(channel index), value: stream index

        # sort the streams by lowest channel index
        for stream_index, stream in enumerate(self.streams):
            stream_channels = stream.channel_index_list[:stream.channel_count]
            d[min(stream_channels)] = stream_index

        for _sort_key, stream_index in sorted(d.items()):
            stream = self.streams[stream_index]
            stream_channels = stream.channel_index_list[:stream.channel_count]

            if not stream_channels:
                continue

            # create the radio button
            radio_button = QtWidgets.QRadioButton(
                "Stream {}".format(stream_index), parent=self)
            self.stream_buttons.append(radio_button)
            self.verticalLayout.addWidget(radio_button)

            group_widget = QtWidgets.QWidget(parent=self)
            self.widgets.append(group_widget)
            layout = QtWidgets.QVBoxLayout(group_widget)

            channel_group = {}
            self.channel_groups.append(channel_group)

            # create the channel check box group
            for channel_index in sorted(stream_channels):
                channel = self.channels[channel_index]
                channel_name = channel.name.decode("utf-8")
                check_box = QtWidgets.QCheckBox(channel_name,
                                                parent=group_widget)
                channel_group[channel_index] = check_box
                check_box.setChecked(False)
                layout.addWidget(check_box)

            self.verticalLayout.addWidget(group_widget)
            group_widget.setEnabled(False)

            # connect the triggered signal to the enabled slot
            radio_button.toggled.connect(group_widget.setEnabled)

        self.stream_buttons[0].setChecked(True)

    def get_channel_list(self):
        for i, stream_button in enumerate(self.stream_buttons):
            if stream_button.isChecked():
                indexes = set()
                for ch_index, check_box in self.channel_groups[i].items():
                    if check_box.isChecked():
                        indexes.add(ch_index)
                return sorted(indexes)

        return []  # error condition, shouldn't happen



import logging

from PySide2 import QtGui, QtWidgets

import asphodel

from .ui_info_dialog import Ui_InfoDialog

logger = logging.getLogger(__name__)


class InfoDialog(QtWidgets.QDialog, Ui_InfoDialog):
    def __init__(self, device_info, parent=None):
        super().__init__(parent)

        self.device_info = device_info

        self.setupUi(self)

        self.font = QtGui.QFont("Monospace", 10)
        self.font.setStyleHint(QtGui.QFont.TypeWriter)
        self.plainTextEdit.setFont(self.font)

        s = self.create_device_info_string()
        self.plainTextEdit.setPlainText(s)

    def create_device_info_string(self):
        d = self.device_info
        s = ""

        s += "Serial Number: {}\n".format(d['serial_number'])
        s += "User Tag 1: {}\n".format(d['user_tag_1'])
        s += "User Tag 2: {}\n".format(d['user_tag_2'])
        try:
            s += "Location String: {}\n".format(d['location_string'])
        except KeyError:
            pass  # file is too old to contain location string
        s += "Max Outgoing Param Len: {}\n".format(
            d['max_outgoing_param_length'])
        s += "Max Incoming Param Len: {}\n".format(
            d['max_incoming_param_length'])
        s += "Stream Packet Length: {}\n".format(d['stream_packet_length'])
        s += "Protocol Version: {}\n".format(d['protocol_version'])
        s += "Board Info: {} rev {}\n".format(*d['board_info'])
        s += "Build Info: {}\n".format(d['build_info'])
        s += "Build Date: {}\n".format(d['build_date'])
        s += "Chip Family: {}\n".format(d['chip_family'])
        s += "Chip Model: {}\n".format(d['chip_model'])
        s += "Chip ID: {}\n".format(d['chip_id'])
        s += "Tag Locations: {}\n".format(
            ", ".join(map(lambda x: str(tuple(x)), d['tag_locations'])))
        s += "Bootloader Info: {}\n".format(d['bootloader_info'])
        s += "\n"
        s += "Library Protocol Version: {}\n".format(
            d['library_protocol_version'])
        s += "Library Build Info: {}\n".format(d['library_build_info'])
        s += "Library Build Date: {}\n".format(d['library_build_date'])
        s += "\n"
        s += "Stream Filler Bits: {}\n".format(d['stream_filler_bits'])
        s += "Stream ID Bits: {}\n".format(d['stream_id_bits'])
        s += "\n"
        s += "Streams\n"

        for i, stream in enumerate(d['streams']):
            s += "  Stream {}\n".format(i)
            channels = stream.channel_index_list[:stream.channel_count]
            s += "    channels: [{}]\n".format(", ".join(map(str, channels)))
            s += "    filler_bits={}, counter_bits={}\n".format(
                stream.filler_bits, stream.counter_bits)
            s += "    rate={:g}, rate_error={:g}%\n".format(
                stream.rate, stream.rate_error * 100.0)
            s += "    warm_up_delay={:g}\n".format(stream.warm_up_delay)
            try:
                rate_info = d['stream_rate_info'][i]
                if rate_info is None or not rate_info.available:
                    s += "    no rate channel\n"
                else:
                    s += "    rate_channel={}, rate_invert={}\n".format(
                        rate_info.channel_index, rate_info.invert)
                    s += "    rate_scale={}, rate_offset={}\n".format(
                        rate_info.scale, rate_info.offset)
            except IndexError:
                # doesn't have rate inforamiton for this stream
                pass

        s += "\n"
        s += "Channels\n"

        for i, channel in enumerate(d['channels']):
            s += "  Channel {}\n".format(i)
            s += "    name: {}\n".format(channel.name.decode("utf-8"))

            stream_ids = []
            for stream_id, stream in enumerate(d['streams']):
                channels = stream.channel_index_list[:stream.channel_count]
                if i in channels:
                    stream_ids.append(stream_id)
            if len(stream_ids) == 1:
                stream_id = stream_ids[0]
                s += "    stream: {}\n".format(stream_id)
                sampling_rate = d['streams'][stream_id].rate * channel.samples
                s += "    rate={:g}\n".format(sampling_rate)
            elif len(stream_ids) == 1:
                s += "    stream: None\n"
            else:
                s += "    streams: [{}]\n".format(", ".join(stream_ids))

            try:
                t_str = asphodel.channel_type_names[channel.channel_type]
                channel_type_str = "{} ({})" .format(channel.channel_type,
                                                     t_str)
            except IndexError:
                channel_type_str = str(channel.channel_type)
            s += "    channel_type={}\n".format(channel_type_str)

            try:
                t_str = asphodel.unit_type_names[channel.unit_type]
                unit_type_str = "{} ({})" .format(channel.unit_type, t_str)
            except IndexError:
                unit_type_str = str(channel.unit_type)
            s += "    unit_type={}\n".format(unit_type_str)

            s += "    filler_bits={}, data_bits={}\n".format(
                channel.filler_bits, channel.data_bits)
            s += "    samples={}, bits_per_sample={}\n".format(
                channel.samples, channel.bits_per_sample)
            s += "    minimum={:g}, maximum={:g}, resolution={:g}\n".format(
                channel.minimum, channel.maximum, channel.resolution)
            coefficients = channel.coefficients[:channel.coefficients_length]
            coefficients_str = ", ".join(map("{:g}".format, coefficients))
            s += "    coefficients: [{}]\n".format(coefficients_str)
            s += "    Chunk Count: {}\n".format(channel.chunk_count)
            for chunk_id in range(channel.chunk_count):
                chunk_len = channel.chunk_lengths[chunk_id]
                chunk = channel.chunks[chunk_id][:chunk_len]
                chunk_str = ",".join(map("{:02x}".format, chunk))
                s += "      Chunk {}: [{}]\n".format(chunk_id, chunk_str)

        s += "\n"
        s += "Supplies\n"

        for i, (name, supply) in enumerate(d['supplies']):
            s += "  Supply {}\n".format(i)
            s += "    name: {}\n".format(name)

            try:
                t_str = asphodel.unit_type_names[supply.unit_type]
                unit_type_str = "{} ({})" .format(supply.unit_type, t_str)
            except IndexError:
                unit_type_str = str(supply.unit_type)
            s += "    unit_type={}\n".format(unit_type_str)
            s += "    is_battery={}, nominal={}\n".format(supply.is_battery,
                                                          supply.nominal)
            s += "    scale={:g}, offset={:g}\n".format(supply.scale,
                                                        supply.offset)

        s += "\n"
        s += "Control Variables\n"

        for i, (name, ctrl_var, _setting) in enumerate(d['ctrl_vars']):
            s += "  Control Variable {}\n".format(i)
            s += "    name: {}\n".format(name)

            try:
                t_str = asphodel.unit_type_names[ctrl_var.unit_type]
                unit_type_str = "{} ({})" .format(ctrl_var.unit_type, t_str)
            except IndexError:
                unit_type_str = str(ctrl_var.unit_type)
            s += "    unit_type={}\n".format(unit_type_str)
            s += "    minimum={}, maximum={}\n".format(ctrl_var.minimum,
                                                       ctrl_var.maximum)
            s += "    scale={:g}, offset={:g}\n".format(ctrl_var.scale,
                                                        ctrl_var.offset)

        s += "\n"
        s += "NVM\n"

        s += "\n".join(asphodel.format_nvm_data(d['nvm']))

        return s



# Copyright (c) 2018 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import builtins
import gc
import logging
import os.path

import numpy
from PySide2 import QtCore, QtGui, QtWidgets
import matplotlib

from . import mondo_rc  # needed for the icon

logger = logging.getLogger(__name__)


ignored_types = [i for i in builtins.__dict__.values() if isinstance(i, type)]
ignored_types.remove(list)
ignored_types.append(type(None))


def is_ignored_type(value):
    if isinstance(value, QtCore.QObject):
        return True
    for t in ignored_types:
        if type(value) == t:
            return True
    return False


class CommandList(list):
    def __init__(self, setup_commands=None):
        self.plotdata = {}
        self.index = 0

        if not setup_commands:
            self.append("import matplotlib")
            self.append("import numpy")
            self.append("import PySide2")
            self.append("")
        else:
            self.extend(setup_commands)
            self.append("")
            self.append("plotdata = numpy.load(__file__ + '.npz')")

    def get_string(self, value):
        if isinstance(value, numpy.ndarray):
            i = 'array_{}'.format(self.index)
            self.index += 1

            self.plotdata[i] = value
            return "plotdata['{}']".format(i)
        else:
            return repr(value)

    def output(self, scriptfile):
        datafile = scriptfile + ".npz"

        logger.debug("Writing script to {}".format(scriptfile))

        with open(scriptfile, "wt", encoding="utf_8") as f:
            for v in self:
                # hack to change fig.show() to plt.show()
                s = str(v)
                if s == "fig.show()":
                    s = "plt.show()"
                f.write(s)
                f.write("\n")

        logger.debug("Writing script data to {}".format(datafile))

        numpy.savez_compressed(datafile, allow_pickle=False, **self.plotdata)

        logger.debug("Finished script output")


class AttributeCommand:
    def __init__(self, cmd_list, base, attr):
        self.cmd_list = cmd_list
        self.base = base
        self.attr = attr

    def __repr__(self):
        return "{}.{}".format(self.base, self.attr)


class CallCommand:
    def __init__(self, cmd_list, base, func, *args, **kwargs):
        self.cmd_list = cmd_list
        self.base = base
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        a = []
        a.extend(self.cmd_list.get_string(x) for x in self.args)
        for k, v in self.kwargs.items():
            a.append(k + "=" + self.cmd_list.get_string(v))
        param_str = ", ".join(a)
        return "{}.{}({})".format(self.base, self.func, param_str)


class GetItemCommand:
    def __init__(self, cmd_list, base, key):
        self.cmd_list = cmd_list
        self.base = base
        self.key = key

    def __repr__(self):
        return "{}[{}]".format(self.base, self.cmd_list.get_string(self.key))


class SetItemCommand:
    def __init__(self, cmd_list, base, key, value):
        self.cmd_list = cmd_list
        self.base = base
        self.key = key
        self.value = value

    def __repr__(self):
        return "{}[{}] = {}".format(self.base,
                                    self.cmd_list.get_string(self.key),
                                    self.cmd_list.get_string(self.value))


class Wrapper:
    def __init__(self, wrapped, name, cmd_list):
        self._wrapped = wrapped
        self._name = name
        self._cmd_list = cmd_list

    def __getattribute__(self, attr):
        wrapped = object.__getattribute__(self, '_wrapped')
        name = object.__getattribute__(self, '_name')
        cmd_list = object.__getattribute__(self, '_cmd_list')
        orig_attr = wrapped.__getattribute__(attr)

        if callable(orig_attr):
            def wrapped_func(*args, **kwargs):
                cmd = CallCommand(cmd_list, name, attr, *args, **kwargs)
                result = orig_attr(*args, **kwargs)
                if is_ignored_type(result):
                    cmd_list.append(cmd)
                    return result
                else:
                    if len(args) + len(kwargs) > 0:
                        cmd_list.append(cmd)
                    return Wrapper(result, cmd, cmd_list)
            return wrapped_func
        else:
            if is_ignored_type(orig_attr):
                return orig_attr
            else:
                new_name = AttributeCommand(cmd_list, name, attr)
                return Wrapper(orig_attr, new_name, cmd_list)

    def __getitem__(self, key):
        wrapped = object.__getattribute__(self, '_wrapped')
        name = object.__getattribute__(self, '_name')
        cmd_list = object.__getattribute__(self, '_cmd_list')
        new_name = GetItemCommand(cmd_list, name, key)
        value = wrapped.__getitem__(key)
        if is_ignored_type(value):
            return value
        else:
            return Wrapper(value, new_name, cmd_list)

    def __setitem__(self, key, value):
        wrapped = object.__getattribute__(self, '_wrapped')
        name = object.__getattribute__(self, '_name')
        cmd_list = object.__getattribute__(self, '_cmd_list')
        cmd_list.append(SetItemCommand(cmd_list, name, key, value))
        return wrapped.__setitem__(key, value)

    def __len__(self):
        wrapped = object.__getattribute__(self, '_wrapped')
        return wrapped.__len__()

    def __iter__(self):
        wrapped = object.__getattribute__(self, '_wrapped')
        return wrapped.__iter__()


def get_script_file(default_name, parent=None):
    settings = QtCore.QSettings()

    # find the directory from settings
    directory = settings.value("fileSaveDirectory")
    if directory and type(directory) == str:
        if not os.path.isdir(directory):
            directory = None

    if not directory:
        directory = ""

    file_and_dir = os.path.join(directory, default_name)

    # ask the user for the file name
    caption = "Save File"
    file_filter = "Python Script (*.py);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getSaveFileName(parent, caption, file_and_dir,
                                                file_filter)
    output_path = val[0]

    if output_path:
        # save the directory
        output_dir = os.path.dirname(output_path)
        settings.setValue("fileSaveDirectory", output_dir)
        return output_path
    else:
        return None


def add_export_script_action(figure, cmd_list):
    toolbar = figure.canvas.toolbar
    actionText = QtWidgets.QApplication.translate("ExportScript",
                                                  "Export Script")
    action = QtWidgets.QAction(actionText, toolbar)
    icon = QtGui.QIcon()
    icon.addPixmap(QtGui.QPixmap(":/icon/icons/export-script.png"))
    action.setIcon(icon)

    def handle_export():
        base_file = get_script_file("export", parent=toolbar)
        if not base_file:
            return

        cmd_list.output(base_file)

    action.triggered.connect(handle_export)

    toolbar.addAction(action)


def subplots_wrapper(*args, **kwargs):
    cmd_list = CommandList(setup_commands)

    cmd_list.append('')

    cmd_list.append('import matplotlib.pyplot as plt')
    import matplotlib.pyplot as plt

    cmd_list.append('')

    fig, ax = plt.subplots(*args, **kwargs)

    def handle_close(evt):
        fig.clf()
        cmd_list.clear()
        gc.collect()

    fig.canvas.mpl_connect("close_event", handle_close)

    # set the icon (inside the application only; not in the script)
    icon = QtGui.QIcon()
    icon_reader = QtGui.QImageReader(":/mondo.ico")
    while True:
        pixmap = QtGui.QPixmap.fromImage(icon_reader.read())
        icon.addPixmap(pixmap)
        if not icon_reader.jumpToNextImage():
            break
    fig.canvas.manager.window.setWindowIcon(icon)

    fig_wrapper = Wrapper(fig, "fig", cmd_list)
    add_export_script_action(fig, cmd_list)

    if isinstance(ax, numpy.ndarray):
        ax_array = ax
        cmd_list.append(CallCommand(
            cmd_list, "fig, ax_array = plt", "subplots", *args, **kwargs))
        ax_wrapper = numpy.empty(ax.shape, dtype=object)
        for i, ax in numpy.ndenumerate(ax_array):
            ax_name = "ax_array[{}]".format(i)
            ax_wrapper[i] = Wrapper(ax, ax_name, cmd_list)
    else:
        cmd_list.append(CallCommand(
            cmd_list, "fig, ax = plt", "subplots", *args, **kwargs))
        ax_wrapper = Wrapper(ax, "ax", cmd_list)

    return fig_wrapper, ax_wrapper


setup_commands = CommandList()
matplotlib_wrapper = Wrapper(matplotlib, 'matplotlib', setup_commands)



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging

from PySide2 import QtWidgets

from .ui_select_subchannels import Ui_SelectSubchannelsDialog

logger = logging.getLogger(__name__)


class SelectSubchannelsDialog(QtWidgets.QDialog, Ui_SelectSubchannelsDialog):
    def __init__(self, names, parent=None):
        super().__init__(parent)

        self.names = names  # [(channel_id, channel_name, subchannel_names)]

        # these are ordered by display order
        self.subchannel_buttons = {}  # (ch_id, subchannel_id): check box

        self.widgets = []  # to prevent premature garbage collection

        self.setupUi(self)

        self.add_buttons()

    def add_buttons(self):
        for channel_id, channel_name, subchannel_names in sorted(self.names):
            if len(subchannel_names) > 1:
                # create a label for the channel
                channel_label = QtWidgets.QLabel(channel_name)
                self.widgets.append(channel_label)
                self.verticalLayout.addWidget(channel_label)

                group_widget = QtWidgets.QWidget(parent=self)
                self.widgets.append(group_widget)
                layout = QtWidgets.QVBoxLayout(group_widget)

                for i, subchannel_name in enumerate(subchannel_names):
                    check_box = QtWidgets.QCheckBox(subchannel_name,
                                                    parent=group_widget)
                    self.subchannel_buttons[(channel_id, i)] = check_box
                    check_box.setChecked(False)
                    layout.addWidget(check_box)

                self.verticalLayout.addWidget(group_widget)
            else:
                # just a single subchannel, no need for the label
                check_box = QtWidgets.QCheckBox(subchannel_names[0])
                self.subchannel_buttons[(channel_id, 0)] = check_box
                check_box.setChecked(False)
                self.widgets.append(check_box)
                self.verticalLayout.addWidget(check_box)

    def get_subchannels_list(self):
        selected_subchannels = []
        for index_tuple in sorted(self.subchannel_buttons.keys()):
            check_box = self.subchannel_buttons[index_tuple]
            if check_box.isChecked():
                selected_subchannels.append(index_tuple)
        return selected_subchannels



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'select_subchannels.ui',
# licensing of 'select_subchannels.ui' applies.
#
# Created: Tue Apr  2 19:36:21 2019
#      by: pyside2-uic  running on PySide2 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_SelectSubchannelsDialog(object):
    def setupUi(self, SelectSubchannelsDialog):
        SelectSubchannelsDialog.setObjectName("SelectSubchannelsDialog")
        SelectSubchannelsDialog.resize(400, 49)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(SelectSubchannelsDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(SelectSubchannelsDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_2.addWidget(self.buttonBox)

        self.retranslateUi(SelectSubchannelsDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), SelectSubchannelsDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), SelectSubchannelsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(SelectSubchannelsDialog)

    def retranslateUi(self, SelectSubchannelsDialog):
        SelectSubchannelsDialog.setWindowTitle(QtWidgets.QApplication.translate("SelectSubchannelsDialog", "Select Subchannels Channels", None, -1))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'setting_viewer.ui',
# licensing of 'setting_viewer.ui' applies.
#
# Created: Tue Mar  3 14:52:37 2020
#      by: pyside2-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_SettingViewerDialog(object):
    def setupUi(self, SettingViewerDialog):
        SettingViewerDialog.setObjectName("SettingViewerDialog")
        SettingViewerDialog.resize(400, 53)
        self.verticalLayout = QtWidgets.QVBoxLayout(SettingViewerDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(SettingViewerDialog)
        self.tabWidget.setObjectName("tabWidget")
        self.verticalLayout.addWidget(self.tabWidget)
        self.buttonBox = QtWidgets.QDialogButtonBox(SettingViewerDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(SettingViewerDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), SettingViewerDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), SettingViewerDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(SettingViewerDialog)

    def retranslateUi(self, SettingViewerDialog):
        SettingViewerDialog.setWindowTitle(QtWidgets.QApplication.translate("SettingViewerDialog", "Device Settings", None, -1))




# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import datetime
from fractions import gcd
import functools
import logging
import os.path

import numpy
from PySide2 import QtCore, QtWidgets

from . import util

logger = logging.getLogger(__name__)


def get_csv_file(default_name, parent=None):
    settings = QtCore.QSettings()

    # find the directory from settings
    directory = settings.value("fileSaveDirectory")
    if directory and type(directory) == str:
        if not os.path.isdir(directory):
            directory = None

    if not directory:
        directory = ""

    file_and_dir = os.path.join(directory, default_name)

    # ask the user for the file name
    caption = "Save File"
    file_filter = "Comma Seperated Value Files (*.csv);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getSaveFileName(parent, caption, file_and_dir,
                                                file_filter)
    output_path = val[0]

    if output_path:
        # save the directory
        output_dir = os.path.dirname(output_path)
        settings.setValue("fileSaveDirectory", output_dir)
        return output_path
    else:
        return None


def do_downsample(array, downsample_factor):
    pad_size = (downsample_factor - (array.size % downsample_factor))
    if pad_size == downsample_factor:
        pad_size = 0
    padded_array = numpy.append(array, numpy.zeros(pad_size) * numpy.nan)
    reshaped = padded_array.reshape((-1, downsample_factor))
    result = numpy.nanmean(reshaped, axis=1)
    return result


def csv_export(parent=None, downsample=False):
    ret = util.load_batch(parent=parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    if downsample:
        downsample_factor, ok = QtWidgets.QInputDialog.getInt(
            parent, "Downsample Factor", "Input downsample factor",
            value=2, minValue=2, maxValue=2147483647)
        if not ok:
            return
    else:
        downsample_factor = 1

    channel_indexes = util.choose_synchronous_channels(header, parent)
    if channel_indexes is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, channel_indexes,
                                   parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, once=True, parent=parent)
    if not ret:
        return  # user cancelled

    sequence_info, unit_info = util.sequence_data(batch_info, chunk=True,
                                                  parent=parent)

    def lcm(iterable):
        return functools.reduce(lambda x, y: x * y // gcd(x, y), iterable)

    samples_lcm = lcm(header['channels'][i].samples for i in channel_indexes)

    # make sure channels all have the same number of gaps
    chunk_count = None
    for sequence in sequence_info.values():
        if chunk_count is None:
            chunk_count = len(sequence)
        else:
            if len(sequence) != chunk_count:
                message = "Channels don't share lost packets!"
                logger.error(message)
                QtWidgets.QMessageBox.critical(parent, "Error", message)
                return

    all_data = []  # holds tuples of (data, first_time)

    for chunk_index in range(chunk_count):
        data_list = []

        first_time = None
        first_data_length = None

        for channel_index in sorted(channel_indexes):
            channel = header['channels'][channel_index]
            sequence = sequence_info[channel_index]
            time, data, _start, _end = sequence[chunk_index]

            expand_factor = samples_lcm // channel.samples
            expanded_data = numpy.repeat(data, expand_factor, axis=0)

            if downsample:
                expanded_data = numpy.apply_along_axis(
                    do_downsample, 0, expanded_data, downsample_factor)

            if first_time is None:
                first_time = time[0]
                first_data_length = len(expanded_data)
            else:
                if first_time != time[0]:
                    message = "Mismatch in channel times!"
                    logger.error(message)
                    QtWidgets.QMessageBox.critical(parent, "Error", message)
                    return
                if first_data_length != len(expanded_data):
                    message = "Mismatch in channel data sizes!"
                    logger.error(message)
                    QtWidgets.QMessageBox.critical(parent, "Error", message)
                    return

            data_list.append(expanded_data)

        all_data.append((numpy.concatenate(data_list, axis=1), first_time))

    first_timestamp = sequence_info[channel_indexes[0]][0][2]
    dt = datetime.datetime.fromtimestamp(first_timestamp,
                                         datetime.timezone.utc)
    default_name = "{}_{}.csv".format(dt.strftime("%Y_%m_%d_%H.%M"),
                                      header['serial_number'])

    stream = batch_info[channel_indexes[0]].stream
    interval = downsample_factor / (stream.rate * samples_lcm)

    row_count = sum(len(data) for data, _time in all_data)
    position = 0

    filename = get_csv_file(default_name)
    if filename is None:
        return

    window_flags = (QtCore.Qt.MSWindowsFixedSizeDialogHint)
    progress = QtWidgets.QProgressDialog("Writing CSV...", "Abort", 0,
                                         row_count, parent, window_flags)
    progress.setWindowModality(QtCore.Qt.WindowModal)
    progress.forceShow()

    with open(filename, "w") as f:
        # write the CSV header
        f.write("Time (s)")
        for channel_index in sorted(channel_indexes):
            channel_data = batch_info[channel_index]
            unit_str = unit_info[channel_index]['ascii']
            for name in channel_data.channel_decoder.subchannel_names:
                if unit_str:
                    f.write(", {} ({})".format(name, unit_str))
                else:
                    f.write(", {}".format(name))
        f.write("\n")

        for data, first_time in all_data:
            for i, row in enumerate(data):
                # print the time
                t = first_time + i * interval
                f.write("{}".format(t))

                for value in row:
                    f.write(", {}".format(value))
                f.write("\n")

                position += 1
                progress.setValue(position)
                if position % 100 == 0:
                    QtCore.QCoreApplication.processEvents()
                    if progress.wasCanceled():
                        message = "Aborting {}!".format(
                            os.path.basename(filename))
                        logger.info(message)
                        return

    message = "Finished writing {}!".format(os.path.basename(filename))
    logger.info(message)
    QtWidgets.QMessageBox.information(parent, "Written", message)



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging

from PySide2 import QtCore, QtWidgets

from .ui_datetime_subset import Ui_DateTimeSubsetDialog

logger = logging.getLogger(__name__)


class DateTimeSubsetDialog(QtWidgets.QDialog, Ui_DateTimeSubsetDialog):
    def __init__(self, start_qdt, end_qdt, parent=None):
        super().__init__(parent, QtCore.Qt.MSWindowsFixedSizeDialogHint)

        self.start_qdt = start_qdt
        self.end_qdt = end_qdt

        self.setupUi(self)

        start_py = start_qdt.toPython()
        end_py = end_qdt.toPython()
        if end_py >= start_py:
            duration = int((end_py - start_py).total_seconds())
        else:
            duration = 0
            self.useSubset.setEnabled(False)

        # write the all data labels
        self.allStart.setText(self.startDateTime.textFromDateTime(start_qdt))
        self.allEnd.setText(self.endDateTime.textFromDateTime(end_qdt))
        self.allDuration.setText("{} s".format(duration))

        self.startDateTime.setDateTime(start_qdt)
        self.startDateTime.setWrapping(True)
        self.endDateTime.setDateTime(end_qdt)
        self.endDateTime.setWrapping(True)

        self.startSeconds.setMinimum(0)
        self.startSeconds.setMaximum(duration)
        self.startSeconds.setValue(0)
        self.endSeconds.setMinimum(-duration)
        self.endSeconds.setMaximum(0)
        self.endSeconds.setValue(0)
        self.duration.setMinimum(0)
        self.duration.setMaximum(duration)
        self.duration.setValue(duration)

        self.startDateTime.setCurrentSection(
            QtWidgets.QDateTimeEdit.SecondSection)
        self.endDateTime.setCurrentSection(
            QtWidgets.QDateTimeEdit.SecondSection)

        self.startDateTime.dateTimeChanged.connect(self.start_changed)
        self.startDateTime.editingFinished.connect(self.start_editing_finished)
        self.endDateTime.dateTimeChanged.connect(self.end_changed)
        self.endDateTime.editingFinished.connect(self.end_editing_finished)

        self.startSeconds.valueChanged.connect(self.start_seconds_changed)
        self.endSeconds.valueChanged.connect(self.end_seconds_changed)
        self.duration.valueChanged.connect(self.duration_changed)

        self.layout().setSizeConstraint(QtWidgets.QLayout.SetFixedSize)

    def should_use_all(self):
        return self.useAllData.isChecked()

    def trim_datetime_to_range(self, value):
        if value < self.start_qdt:
            return self.start_qdt
        elif value > self.end_qdt:
            return self.end_qdt
        else:
            return value

    def datetime_in_range(self, value):
        return ((self.start_qdt <= value) and (value <= self.end_qdt))

    def get_subset(self):
        start_qdt = self.trim_datetime_to_range(self.startDateTime.dateTime())
        end_qdt = self.trim_datetime_to_range(self.endDateTime.dateTime())
        return (start_qdt, end_qdt)

    def start_changed(self, new_date_time):
        self.update_duration(moving_start=True)

        if self.datetime_in_range(new_date_time):
            if self.endDateTime.dateTime() < new_date_time:
                # start is later than end
                self.startDateTime.setStyleSheet(
                    "* { background-color: yellow }")
            else:
                # valid
                self.startDateTime.setStyleSheet("")
        else:
            # outside the bounds
            self.startDateTime.setStyleSheet("* { background-color: red }")

    def end_changed(self, new_date_time):
        self.update_duration(moving_start=False)

        if self.datetime_in_range(new_date_time):
            if new_date_time < self.startDateTime.dateTime():
                # end is before start
                self.endDateTime.setStyleSheet(
                    "* { background-color: yellow }")
            else:
                # valid
                self.endDateTime.setStyleSheet("")
        else:
            # outside the bounds
            self.endDateTime.setStyleSheet("* { background-color: red }")

    def start_editing_finished(self):
        if self.datetime_in_range(self.startDateTime.dateTime()):
            if self.endDateTime.dateTime() <= self.startDateTime.dateTime():
                self.endDateTime.setDateTime(self.startDateTime.dateTime())

                # trigger revalidation
                self.startDateTime.setDateTime(self.startDateTime.dateTime())
        else:
            self.startDateTime.setDateTime(self.start_qdt)

    def end_editing_finished(self):
        if self.datetime_in_range(self.endDateTime.dateTime()):
            if self.endDateTime.dateTime() <= self.startDateTime.dateTime():
                self.startDateTime.setDateTime(self.endDateTime.dateTime())

                # trigger revalidation
                self.endDateTime.setDateTime(self.endDateTime.dateTime())
        else:
            self.endDateTime.setDateTime(self.end_qdt)

    def update_duration(self, moving_start=True):
        start_qdt = self.startDateTime.dateTime()
        end_qdt = self.endDateTime.dateTime()

        if start_qdt < self.start_qdt:
            # clip to beginning
            start_py = self.start_qdt.toPython()
        elif self.end_qdt < start_qdt:
            # clip to end
            start_py = self.end_qdt.toPython()
        else:
            # ok
            start_py = start_qdt.toPython()

        if end_qdt < self.start_qdt:
            # clip to beginning
            end_py = self.start_qdt.toPython()
        elif self.end_qdt < end_qdt:
            # clip to end
            end_py = self.end_qdt.toPython()
        else:
            # ok
            end_py = end_qdt.toPython()

        if end_py < start_py:
            # overlapping
            if moving_start:
                end_py = start_py
            else:
                start_py = end_py

        start_seconds = (start_py - self.start_qdt.toPython()).total_seconds()
        self.startSeconds.setValue(start_seconds)
        end_seconds = (end_py - self.end_qdt.toPython()).total_seconds()
        self.endSeconds.setValue(end_seconds)

        self.duration.setValue((end_py - start_py).total_seconds())

    def start_seconds_changed(self, junk=None):
        new_start = self.start_qdt.addSecs(self.startSeconds.value())
        if new_start > self.endDateTime.dateTime():
            self.endDateTime.setDateTime(new_start)
        self.startDateTime.setDateTime(new_start)

    def end_seconds_changed(self, junk=None):
        new_end = self.end_qdt.addSecs(self.endSeconds.value())
        if new_end < self.startDateTime.dateTime():
            self.startDateTime.setDateTime(new_end)
        self.endDateTime.setDateTime(new_end)

    def duration_changed(self, junk=None):
        duration = self.duration.value()
        current_start = self.startDateTime.dateTime().toPython()
        end_py = self.end_qdt.toPython()
        headroom = int((end_py - current_start).total_seconds())

        if duration <= headroom:
            # just move the end
            self.endSeconds.setValue(duration - headroom)
        else:
            # need to move the start too
            self.endDateTime.setDateTime(self.end_qdt)
            self.startDateTime.setDateTime(self.end_qdt.addSecs(-duration))



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import math

import numpy
from PySide2 import QtCore, QtWidgets
import scipy.signal.windows

from .ui_psd_options import Ui_PSDOptionsWidget

logger = logging.getLogger(__name__)


class PSDOptionsWidget(QtWidgets.QWidget, Ui_PSDOptionsWidget):
    def __init__(self, chunks_list, sampling_rate, parent=None):
        super().__init__(parent)

        self.chunks_list = chunks_list
        self.sampling_rate = sampling_rate

        self.setupUi(self)

        self.fill_combo_boxes()

        self.changing_overlap = False

        self.fftPoints.currentIndexChanged.connect(self.fft_points_changed)
        self.overlapPercent.valueChanged.connect(self.overlap_percent_changed)
        self.overlapPoints.valueChanged.connect(self.overlap_points_changed)

        self.fft_points_changed()
        self.overlapPercent.setValue(50.0)

        self.layout().setSizeConstraint(QtWidgets.QLayout.SetFixedSize)

    def fill_combo_boxes(self):
        max_lengths = []
        for chunks in self.chunks_list:
            max_lengths.append(max(len(c[0]) for c in chunks))
        max_length = min(max_lengths)

        max_pow_2 = math.floor(math.log2(max_length))
        min_pow_2 = 8  # 256 points

        self.fft_points_by_index = []
        for i in range(min_pow_2, max_pow_2 + 1):
            points = 2 ** i
            s = "{} (2^{})".format(points, i)
            self.fft_points_by_index.append(points)
            self.fftPoints.addItem(s)

        default_pow_2 = max(min_pow_2, max_pow_2 - 3)
        default_index = default_pow_2 - min_pow_2
        self.fftPoints.setCurrentIndex(default_index)

        # populate window functions
        self.window_functions = [("Hann", self.hann_window),
                                 ("Hamming", self.hamming_window),
                                 ("Flat Top", self.flattop_window),
                                 ("None (Uniform)", self.uniform_window)]
        for key, _value in self.window_functions:
            self.windowFunction.addItem(key)
        self.windowFunction.setCurrentIndex(0)

        self.detrend_options = [("Linear", "linear"),
                                ("Mean", "mean"),
                                ("None", "none")]
        for key, _value in self.detrend_options:
            self.detrendMethod.addItem(key)
        self.detrendMethod.setCurrentIndex(0)

    def update_window_count(self, window_size, overlap_points):
        window_counts = []
        for chunks in self.chunks_list:
            window_count = 0
            for chunk in chunks:
                chunk_size = len(chunk[0])
                window_count += (1 + ((chunk_size - window_size) //
                                      (window_size - overlap_points)))
            window_counts.append(window_count)
        s = ", ".join(str(w) for w in window_counts)
        self.windowCount.setText(s)

    def fft_points_changed(self, junk=None):
        index = self.fftPoints.currentIndex()
        window_size = self.fft_points_by_index[index]

        original_percent = self.overlapPercent.value()

        if self.overlapPoints.value() >= window_size:
            self.overlapPoints.setValue(window_size - 1)
        self.overlapPoints.setMaximum(window_size - 1)
        self.overlapPercent.setMaximum(100 * (window_size - 1) / window_size)

        # update the percentage (use existing)
        self.changing_overlap = True
        self.overlapPercent.setValue(original_percent)
        self.changing_overlap = False
        self.overlap_percent_changed()

        window_duration = window_size / self.sampling_rate
        self.duration.setText("{:.3f} s".format(window_duration))

        frequency_resolution = self.sampling_rate / window_size
        self.resolution.setText("{:.3f} Hz".format(frequency_resolution))

    def overlap_percent_changed(self, junk=None):
        if self.changing_overlap:
            return
        try:
            self.changing_overlap = True

            index = self.fftPoints.currentIndex()
            window_size = self.fft_points_by_index[index]

            percent = self.overlapPercent.value()
            points = math.floor(window_size * percent / 100)
            if points >= window_size:
                points = window_size - 1
            self.overlapPoints.setValue(points)

            self.update_window_count(window_size, points)
        finally:
            self.changing_overlap = False

    def overlap_points_changed(self, junk=None):
        if self.changing_overlap:
            return
        try:
            self.changing_overlap = True

            index = self.fftPoints.currentIndex()
            window_size = self.fft_points_by_index[index]

            points = self.overlapPoints.value()
            percent = points / window_size * 100
            self.overlapPercent.setValue(percent)

            self.update_window_count(window_size, points)
        finally:
            self.changing_overlap = False

    def hann_window(self, x):
        return scipy.signal.windows.hann(len(x), False) * x

    def hamming_window(self, x):
        return scipy.signal.windows.hamming(len(x), False) * x

    def flattop_window(self, x):
        return scipy.signal.windows.flattop(len(x), False) * x

    def uniform_window(self, x):
        return x

    def get_options(self):
        options = {"Fs": self.sampling_rate}

        index = self.fftPoints.currentIndex()
        window_size = self.fft_points_by_index[index]
        options['NFFT'] = window_size

        options['noverlap'] = self.overlapPoints.value()

        index = self.windowFunction.currentIndex()
        window_function = self.window_functions[index][1]
        options['window'] = window_function(numpy.ones(window_size,
                                                       dtype=numpy.double))

        index = self.detrendMethod.currentIndex()
        options['detrend'] = self.detrend_options[index][1]

        return options


class PSDOptionsDialog(QtWidgets.QDialog):
    def __init__(self, chunks_list, sampling_rate, parent=None):
        super().__init__(parent, QtCore.Qt.MSWindowsFixedSizeDialogHint)

        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.psd_options = PSDOptionsWidget(chunks_list, sampling_rate, self)
        self.verticalLayout.addWidget(self.psd_options)
        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel |
                                          QtWidgets.QDialogButtonBox.Ok)
        self.verticalLayout.addWidget(self.buttonBox)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.layout().setSizeConstraint(QtWidgets.QLayout.SetFixedSize)

        self.setWindowTitle(self.tr("PSD Options"))

    def get_options(self):
        return self.psd_options.get_options()


class MultiplePSDOptionsDialog(QtWidgets.QDialog):
    def __init__(self, sections, parent=None):
        # NOTE: WindowTitleHint | WindowSystemMenuHint disables "What's This"
        super().__init__(parent, QtCore.Qt.WindowTitleHint |
                         QtCore.Qt.WindowSystemMenuHint)

        self.psd_options = []
        self.extra_widgets = []  # to prevent garbage collection

        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.scrollArea = QtWidgets.QScrollArea(self)
        self.scrollArea.setVerticalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setFrameShape(QtWidgets.QFrame.NoFrame)

        self.scrollAreaContents = QtWidgets.QWidget()
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.scrollAreaContents)

        for i, (label_text, chunks_list, sampling_rate) in enumerate(sections):
            groupbox = QtWidgets.QGroupBox(self)
            self.extra_widgets.append(groupbox)

            layout = QtWidgets.QVBoxLayout(groupbox)

            label = QtWidgets.QLabel(groupbox)
            label.setText(label_text)
            label.setWordWrap(True)
            self.extra_widgets.append(label)
            layout.addWidget(label)

            spacer = QtWidgets.QSpacerItem(
                0, 20, QtWidgets.QSizePolicy.Minimum,
                QtWidgets.QSizePolicy.MinimumExpanding)
            layout.addItem(spacer)

            widget = PSDOptionsWidget(chunks_list, sampling_rate, groupbox)
            self.psd_options.append(widget)
            layout.addWidget(widget)

            self.horizontalLayout.addWidget(groupbox)

            if i < 3:
                # set the minimum size to 3 complete groups
                size_hint = self.scrollAreaContents.sizeHint()
                horizontal_bar = self.scrollArea.horizontalScrollBar()
                min_bar_height = horizontal_bar.sizeHint().height()
                self.scrollArea.setMinimumHeight(size_hint.height() +
                                                 min_bar_height)
                self.scrollArea.setMinimumWidth(size_hint.width())

        self.scrollArea.setWidget(self.scrollAreaContents)
        self.verticalLayout.addWidget(self.scrollArea)

        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel |
                                          QtWidgets.QDialogButtonBox.Ok)
        self.verticalLayout.addWidget(self.buttonBox)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.setWindowTitle(self.tr("PSD Options"))

    def get_options(self):
        return [w.get_options() for w in self.psd_options]



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# wrapper around "import matplotlib" that captures any commands for later use
from .export_script import matplotlib_wrapper as matplotlib

# import matplotlib and have it use PySide
import PySide2
matplotlib.use('Qt4Agg')
matplotlib.rcParams['agg.path.chunksize'] = 10000
matplotlib.rcParams['figure.constrained_layout.use'] = True
matplotlib.rcParams['image.interpolation'] = 'bilinear'
matplotlib.rcParams['image.resample'] = False

try:
    from .version import version as __version__
except ImportError:
    __version__ = "UNKNOWN"


if __name__ == '__main__':
    from . import __main__
    __main__.main()



import io
import os.path
import subprocess

import PySide2.scripts


def create_py_from_qrc(qrc_file):
    base = os.path.dirname(os.path.realpath(PySide2.scripts.__file__))
    command = os.path.abspath(os.path.join(base, "..", 'pyside2-rcc'))
    s = subprocess.check_output([command, os.path.abspath(qrc_file)],
                                encoding='utf-8')
    return s


def compare_and_update(filename, new_contents):
    with open(filename, 'r') as f:
        old_lines = f.readlines()
    new_lines = new_contents.splitlines(keepends=True)

    new_uncommented = [l for l in new_lines if not l.startswith("#")]
    old_uncommented = [l for l in old_lines if not l.startswith("#")]

    if new_uncommented == old_uncommented:
        print("No Change: {}".format(os.path.basename(filename)))
        return
    else:
        with open(filename, 'w') as f:
            f.write(new_contents)
        print("Updated: {}".format(os.path.basename(filename)))


def update_qrc(qrc_file):
    # pull off the .qrc
    root, _ext = os.path.splitext(qrc_file)
    if _ext != ".qrc":
        root = qrc_file
    
    # add _rc.py to the end
    output_file = root + "_rc.py"

    new_contents = create_py_from_qrc(qrc_file)
    compare_and_update(output_file, new_contents)


def main():
    for root, dirs, files in os.walk("."):
        # skip hidden files and folders
        files = [f for f in files if not f[0] == '.']
        dirs[:] = [d for d in dirs if not d[0] == '.']

        # skip build directory
        if 'build' in dirs:
            dirs.remove('build')

        for file in files:
            if file.endswith('.qrc'):
                update_qrc(os.path.join(root, file))


if __name__ == "__main__":
    main()



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import binascii
import bisect
import datetime
import json
import logging
import lzma
import operator
import os.path
import struct

import numpy
from PySide2 import QtCore, QtWidgets

import asphodel
import hyperborea.unit_preferences

from ..datetime_subset import DateTimeSubsetDialog
from ..psd_options import PSDOptionsDialog, MultiplePSDOptionsDialog
from ..select_subchannels import SelectSubchannelsDialog
from ..select_synchronous import SelectSynchronousDialog

logger = logging.getLogger(__name__)


def get_packdata_file(parent=None):
    settings = QtCore.QSettings()

    # find the directory from settings
    directory = settings.value("fileOpenDirectory")
    if directory and type(directory) == str:
        if not os.path.isdir(directory):
            directory = None

    if not directory:
        directory = ""

    # ask the user for the file name
    caption = "Open File"
    file_filter = "Packed Data Files (*.apd);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getOpenFileName(parent, caption, directory,
                                                file_filter)
    output_path = val[0]

    if output_path:
        # save the directory
        output_dir = os.path.dirname(output_path)
        settings.setValue("fileOpenDirectory", output_dir)
        return output_path
    else:
        return None


def get_packdata_files(parent=None):
    settings = QtCore.QSettings()

    # find the directory from settings
    directory = settings.value("fileOpenDirectory")
    if directory and type(directory) == str:
        if not os.path.isdir(directory):
            directory = None

    if not directory:
        directory = ""

    # ask the user for the file name
    caption = "Open Files"
    file_filter = "Packed Data Files (*.apd);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getOpenFileNames(parent, caption, directory,
                                                 file_filter)
    files = val[0]

    if len(files) > 0:
        # save the directory (use 1st file for simplicity)
        output_dir = os.path.dirname(files[0])
        settings.setValue("fileOpenDirectory", output_dir)
        return sorted(files)
    else:
        return None


def decode_header(header_bytes, parent=None):
    try:
        header_str = header_bytes.decode("UTF-8")
        header = json.loads(header_str)
    except:
        message = "Could not parse file header!"
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return

    try:
        # convert the JSON info back into an actual Asphodel structure
        all_streams = [asphodel.AsphodelStreamInfo.from_json_obj(s)
                       for s in header['streams']]
        all_channels = [asphodel.AsphodelChannelInfo.from_json_obj(c)
                        for c in header['channels']]

        header['streams'] = all_streams
        header['channels'] = all_channels

        # stream rate info
        stream_rate_info = []
        for values in header.get('stream_rate_info', []):
            # fix floats getting converted to strings in older files
            values = [float(v) if isinstance(v, str) else v for v in values]

            if values is not None:
                stream_rate_info.append(asphodel.StreamRateInfo(*values))
            else:
                stream_rate_info.append(None)
        header['stream_rate_info'] = stream_rate_info

        # supplies
        supplies = []
        for name, values in header.get('supplies', []):
            # fix floats getting converted to strings in older files
            values = [float(v) if isinstance(v, str) else v for v in values]

            supplies.append((name, asphodel.SupplyInfo(*values)))
        header['supplies'] = supplies

        # control variables
        ctrl_vars = []
        for name, values, setting in header.get('ctrl_vars', []):
            # fix floats getting converted to strings in older files
            values = [float(v) if isinstance(v, str) else v for v in values]

            ctrl_vars.append((name, asphodel.CtrlVarInfo(*values), setting))
        header['ctrl_vars'] = ctrl_vars

        # nvm
        header['nvm'] = binascii.a2b_hex(header['nvm'])

        # custom enums: need to convert keys back from strings to ints
        custom_enums = {int(k): v for k, v in header['custom_enums'].items()}
        header['custom_enums'] = custom_enums

        # settings
        settings = []
        for setting_str in header['settings']:
            try:
                setting = asphodel.AsphodelSettingInfo.from_str(setting_str)
            except:
                setting = None
            settings.append(setting)
        header['settings'] = settings

    except:
        message = "Unknown error in file header!"
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return

    return header


def load_single_file(parent=None):
    """
    returns (file_infos, header) where file_infos is a sequence of tuples
    containing (filename, dt).
    * header is the dictionary loaded from the file's JSON data, with
      appropriate conversions applied to Asphodel struct data.
    * filename is the absolute path to the file location.
    * dt is the UTC datetime of the first packet in the file.
    """

    filename = get_packdata_file(parent)
    if filename is None:
        return None  # user cancelled

    try:
        with lzma.LZMAFile(filename, "rb") as f:
            # read the header
            header_leader = struct.unpack(">dI", f.read(12))
            _header_dt = datetime.datetime.fromtimestamp(
                header_leader[0], datetime.timezone.utc)
            header_bytes = f.read(header_leader[1])

            if len(header_bytes) == 0:
                message = "Empty header in {}!".format(filename)
                logger.error(message)
                QtWidgets.QMessageBox.critical(parent, "Error", message)
                return None  # error

            # read the first packet's datetime
            first_packet_timestamp = struct.unpack(">d", f.read(8))[0]
            first_packet_dt = datetime.datetime.fromtimestamp(
                first_packet_timestamp, datetime.timezone.utc)

            header = decode_header(header_bytes, parent=parent)

            return ([(filename, first_packet_dt)], header)

    except:
        message = "Could not read header on {}!".format(filename)
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return None  # error


def load_batch(parent=None):
    """
    returns (file_infos, header) where file_infos is a sequence of tuples
    containing (filename, dt).
    * header is the dictionary loaded from the file's JSON data, with
      appropriate conversions applied to Asphodel struct data.
    * filename is the absolute path to the file location.
    * dt is the UTC datetime of the first packet in the file.
    """

    first_file = True

    loaded_files = []

    file_infos = []

    while True:
        files = get_packdata_files(parent)
        if files is None or len(files) == 0:
            # user cancelled
            if first_file:
                return None
            else:
                files = []

        for filename in files:
            if filename in loaded_files:
                message = "File {} already loaded! Skipping.".format(filename)
                logger.info(message)
                QtWidgets.QMessageBox.information(parent, "Already Loaded",
                                                  message)
                continue

            try:
                with lzma.LZMAFile(filename, "rb") as f:
                    # read the header
                    header_leader = struct.unpack(">dI", f.read(12))
                    header_dt = datetime.datetime.fromtimestamp(
                        header_leader[0], datetime.timezone.utc)
                    header_bytes = f.read(header_leader[1])

                    if len(header_bytes) == 0:
                        message = "Empty header in {}!".format(filename)
                        logger.error(message)
                        QtWidgets.QMessageBox.critical(
                            parent, "Error", message)
                        return None  # error

                    # read the first packet's datetime
                    first_packet_timestamp = struct.unpack(">d", f.read(8))[0]
                    first_packet_dt = datetime.datetime.fromtimestamp(
                        first_packet_timestamp, datetime.timezone.utc)

                    if first_file:
                        first_file = False
                        first_header_bytes = header_bytes
                        first_header_dt = header_dt

                        header = decode_header(header_bytes, parent=parent)
                        if not header:
                            return  # error message already displayed
                    else:
                        if (first_header_bytes != header_bytes or
                                first_header_dt != header_dt):
                            # error
                            s = ("Headers do not match on {}!\n"
                                 "\n"
                                 "Files must come from the same session.")
                            message = s.format(filename)
                            logger.error(message)
                            QtWidgets.QMessageBox.critical(parent, "Error",
                                                           message)
                            return None  # error

                    file_infos.append((filename, first_packet_dt))
                    loaded_files.append(filename)

            except:
                message = "Could not read header on {}!".format(filename)
                logger.exception(message)
                QtWidgets.QMessageBox.critical(parent, "Error", message)
                return None  # error

        # ask the user if they want to load more files
        ret = QtWidgets.QMessageBox.question(
            parent, "More Files?", "Load more files for this batch?",
            buttons=(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No),
            default=QtWidgets.QMessageBox.No)

        if ret != QtWidgets.QMessageBox.Yes:
            break

    return (file_infos, header)


def choose_channel(header, parent=None):
    """
    returns a channel index selected by the user.
    """
    channel_indexes = set()
    for stream_id in header['streams_to_activate']:
        stream = header['streams'][stream_id]
        for index in stream.channel_index_list[0:stream.channel_count]:
            channel_indexes.add(index)

    channel_names = []
    name_dict = {}
    for channel_index in sorted(channel_indexes):
        channel = header['channels'][channel_index]
        channel_name = channel.name[0:channel.name_length].decode("UTF-8")
        channel_names.append(channel_name)
        name_dict[channel_name] = channel_index

    value, ok = QtWidgets.QInputDialog.getItem(parent, "Select Channel",
                                               "Select Channel", channel_names,
                                               0, editable=False)
    if not ok:
        return  # user cancelled

    return name_dict[value]


def choose_synchronous_channels(header, parent=None):
    """
    returns a sequence of channel indexes selected by the user, from a single
    stream.
    """

    dialog = SelectSynchronousDialog(header['streams'], header['channels'],
                                     parent=parent)
    ret = dialog.exec_()
    if ret == 0:
        return  # user cancelled

    channel_list = dialog.get_channel_list()

    if len(channel_list) == 0:
        return None  # user didn't select any channels; treat like a cancel
    else:
        return channel_list


class ChannelData:
    def __init__(self, stream, channel, channel_decoder):
        self.stream = stream
        self.channel = channel
        self.channel_decoder = channel_decoder

        self.samples = self.channel.samples
        self.subchannels = self.channel_decoder.subchannels

        if self.samples == 0 or self.subchannels == 0 or stream.rate == 0:
            raise ValueError("Invalid channel configuration")

        self.last_counter = -1

        self.lost_packet_list = []
        self.lost_packet_file_boundary_list = []

        self.file_boundary = False
        self.next_timestamp = None

        self.allocated = 1000
        self.length = 0
        self.data = numpy.empty((self.allocated * self.samples,
                                 self.subchannels), dtype=numpy.double)
        self.indexes = numpy.empty((self.allocated,), dtype=numpy.uint64)
        self.timestamps = numpy.empty((self.allocated,), dtype=numpy.double)

    def set_file_boundary(self):
        self.file_boundary = True

    def set_next_timestamp(self, timestamp):
        self.next_timestamp = timestamp

    def trim(self):
        self.allocated = self.length
        self.data = self.data[0:self.length * self.samples]
        self.indexes = self.indexes[0:self.length]
        self.timestamps = self.timestamps[0:self.length]

    def decode_callback(self, counter, data, samples, subchannels):
        if self.samples != samples:
            raise ValueError("Bad sample count in callback!")

        if self.subchannels != subchannels:
            raise ValueError("Bad subchannel count in callback!")

        if self.last_counter + 1 != counter and self.length != 0:
            # NOTE: first data point being non-zero doesn't count as a loss
            lost_tuple = (self.last_counter, counter, self.length)
            self.lost_packet_list.append(lost_tuple)
            if self.file_boundary:
                self.lost_packet_file_boundary_list.append(lost_tuple)

        self.last_counter = counter

        if self.length == self.allocated:
            # increase the size of the arrays

            self.allocated = self.allocated * 2

            new_data = numpy.empty((self.allocated * self.samples,
                                    self.subchannels), dtype=numpy.double)
            new_data[0:self.length * self.samples] = self.data
            self.data = new_data

            new_indexes = numpy.empty((self.allocated,), dtype=numpy.uint64)
            new_indexes[0:self.length] = self.indexes
            self.indexes = new_indexes

            new_timestamps = numpy.empty((self.allocated,), dtype=numpy.double)
            new_timestamps[0:self.length] = self.timestamps
            self.timestamps = new_timestamps

        d = numpy.array(data).reshape(samples, subchannels)

        self.indexes[self.length] = counter
        self.timestamps[self.length] = self.next_timestamp
        self.data[self.length * samples:(self.length + 1) * samples] = d

        self.length += 1

        if self.file_boundary:
            self.file_boundary = False


def decode_batch(file_infos, header, channel_indexes, parent=None):
    """
    returns a map with keys of channel indexes, values of ChannelData
    """
    info_list = []
    remaining_channels = set(channel_indexes)
    for stream_id in header['streams_to_activate']:
        stream = header['streams'][stream_id]
        indexes = stream.channel_index_list[0:stream.channel_count]
        use_stream = False

        for index in indexes:
            if index in remaining_channels:
                use_stream = True
                remaining_channels.discard(index)

        if use_stream:
            channel_list = [header['channels'][ch_id] for ch_id in indexes]
            info_list.append((stream_id, stream, channel_list))

    # create the device decoder
    decoder = asphodel.nativelib.create_device_decoder(
        info_list, header['stream_filler_bits'], header['stream_id_bits'])

    # create and register callbacks
    batch_info = {}
    remaining_channels = set(channel_indexes)
    try:
        for i, stream_decoder in enumerate(decoder.decoders):
            stream = header['streams'][decoder.stream_ids[i]]
            for j, channel_decoder in enumerate(stream_decoder.decoders):
                channel_id = stream_decoder.stream_info.channel_index_list[j]
                if channel_id in remaining_channels:
                    remaining_channels.discard(channel_id)
                else:
                    continue
                channel = header['channels'][channel_id]

                channel_data = ChannelData(stream, channel, channel_decoder)
                batch_info[channel_id] = channel_data
                channel_decoder.set_callback(channel_data.decode_callback)
    except:
        message = "Could not create device decoder!"
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return None  # error

    # sort by first packet datetime, so earlier files get decoded first
    sorted_file_infos = sorted(file_infos, key=operator.itemgetter(1))

    opened_files = []  # list of (filename, rawfile, rawlen, lzmafile)
    total_length = 0

    # open input files and find their lengths
    for filename, _first_packet_dt in sorted_file_infos:
        try:
            # open the compressed file, and figure out how many bytes it has
            rawfile = open(filename, 'rb')
            rawfile.seek(0, os.SEEK_END)
            rawlen = rawfile.tell()
            total_length += rawlen
            rawfile.seek(0, os.SEEK_SET)

            # open the LZMA wrapper object
            lzmafile = lzma.LZMAFile(rawfile, "rb")

            # skip over the header (since it's already been processed)
            header_leader = struct.unpack(">dI", lzmafile.read(12))
            _header_bytes = lzmafile.read(header_leader[1])

            opened_files.append((filename, rawfile, rawlen, lzmafile))
        except:
            message = "Could not open file {}!".format(filename)
            logger.exception(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return None  # error

    packet_leader = struct.Struct(">dI")
    stream_packet_length = header['stream_packet_length']

    # create a progress dialog box
    window_flags = (QtCore.Qt.MSWindowsFixedSizeDialogHint)
    progress = QtWidgets.QProgressDialog("Loading Data...", "Abort", 0,
                                         total_length, parent, window_flags)
    progress.setWindowModality(QtCore.Qt.WindowModal)
    progress.forceShow()

    # make sure it gets at least one value different than total_length
    progress.setValue(0)

    process_countdown = 0
    total_position = 0

    # start processing the files
    for filename, rawfile, rawlen, lzmafile in opened_files:
        # notify channel data about a file boundary
        for channel_data in batch_info.values():
            channel_data.set_file_boundary()

        try:
            while True:
                leader_bytes = lzmafile.read(packet_leader.size)

                if not leader_bytes:
                    break  # file is finished

                leader = packet_leader.unpack(leader_bytes)
                packet_timestamp = leader[0]
                packet_bytes = lzmafile.read(leader[1])

                for channel_data in batch_info.values():
                    channel_data.set_next_timestamp(packet_timestamp)

                for i in range(len(packet_bytes) // stream_packet_length):
                    start = i * stream_packet_length
                    packet = packet_bytes[start:start + stream_packet_length]
                    decoder.decode(packet)

                file_position = rawfile.tell()
                progress.setValue(total_position + file_position)

                # update the window periodically
                if process_countdown == 0:
                    process_countdown = 100
                    QtCore.QCoreApplication.processEvents()
                else:
                    process_countdown -= 1

                # abort if required
                if progress.wasCanceled():
                    return  # user cancelled
        except:
            progress.cancel()
            m = "Error while processing data in file {}!".format(filename)
            logger.exception(m)

            # ask the user if they want to ignore the error
            ret = QtWidgets.QMessageBox.critical(
                parent, "Error", m, buttons=(QtWidgets.QMessageBox.Abort |
                                             QtWidgets.QMessageBox.Ignore),
                default=QtWidgets.QMessageBox.Abort)
            if ret != QtWidgets.QMessageBox.Ignore:
                return  # user aborted
            else:
                break

        total_position += rawlen
        logger.debug("Finished reading {}.".format(filename))

    for channel_data in batch_info.values():
        channel_data.trim()

    return batch_info


def get_datetime_subset(batch_info, parent=None):
    """
    filters data by datetime, provided by the user.
    """
    for _channel_index, channel_data in sorted(batch_info.items()):
        channel_name = channel_data.channel_decoder.channel_name

        d = numpy.diff(channel_data.timestamps)
        if numpy.min(d) < 0.0:
            # timestamps aren't monotonic!
            s = "Timestamps are not monotonic on channel {}! Continue?"
            message = s.format(channel_name)
            logger.warning(message)
            ret = QtWidgets.QMessageBox.warning(
                parent, "Warning", message,
                buttons=(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No),
                default=QtWidgets.QMessageBox.Yes)

            if ret == QtWidgets.QMessageBox.No:
                return None  # user aborted

    starts = [d.timestamps[0] for d in batch_info.values()]
    ends = [d.timestamps[-1] for d in batch_info.values()]

    # NOTE: floor and ceil allow seconds to be rounded neatly
    start = datetime.datetime.fromtimestamp(numpy.floor(min(starts)),
                                            datetime.timezone.utc)
    end = datetime.datetime.fromtimestamp(numpy.ceil(max(ends)),
                                          datetime.timezone.utc)

    start_qdt = QtCore.QDateTime(start)
    end_qdt = QtCore.QDateTime(end)

    dialog = DateTimeSubsetDialog(start_qdt, end_qdt, parent=parent)

    while True:  # allow opening the dialog multiple times
        ret = dialog.exec_()
        if ret == 0:
            return  # user cancelled

        if dialog.should_use_all():
            return batch_info
        else:
            start_qdt, end_qdt = dialog.get_subset()
            utc = datetime.timezone.utc
            start = start_qdt.toPython().replace(tzinfo=utc).timestamp()
            end = end_qdt.toPython().replace(tzinfo=utc).timestamp()

            new_indexes = {}
            some_empty = False

            for channel_index, channel_data in batch_info.items():
                start_index = bisect.bisect_left(channel_data.timestamps,
                                                 start)
                end_index = bisect.bisect_right(channel_data.timestamps, end)

                if start_index >= end_index:
                    some_empty = True

                new_indexes[channel_index] = (start_index, end_index)

            if some_empty:
                message = ("No data is present within this interval on one or "
                           "more channels.")
                logger.warning(message)
                QtWidgets.QMessageBox.warning(parent, "Warning", message)
                continue  # bring the dialog back up

            for channel_index, channel_data in batch_info.items():
                # trim the data
                start_index, end_index = new_indexes[channel_index]
                if start_index >= end_index:
                    # remove possibility of negative lengths
                    start_index = 0
                    end_index = 0

                data_start = start_index * channel_data.samples
                data_end = end_index * channel_data.samples
                channel_data.data = channel_data.data[data_start:data_end]
                channel_data.indexes = channel_data.indexes[start_index:
                                                            end_index]
                channel_data.timestamps = channel_data.timestamps[start_index:
                                                                  end_index]

                def trim_list(lost_packet_list, start_index, end_index):
                    new_lost_packet_list = []
                    for last, current, index in lost_packet_list:
                        # NOTE: bounds are not inclusive, since we can't have
                        # lost data before the start or after the end.
                        if index > start_index and index < end_index:
                            new_lost_packet_list.append((last, current,
                                                         index - start_index))
                    return new_lost_packet_list

                channel_data.lost_packet_list = trim_list(
                    channel_data.lost_packet_list, start_index, end_index)
                channel_data.lost_packet_file_boundary_list = trim_list(
                    channel_data.lost_packet_file_boundary_list, start_index,
                    end_index)

                channel_data.length = end_index - start_index
                channel_data.allocated = channel_data.length

            return batch_info


def warn_about_lost_packets(batch_info, once=False, parent=None):
    """
    Warns the user about problems with the decoded data, such as lost packets
    or gaps between files. Returns True to continue, False if the user aborts.
    """
    for _channel_index, channel_data in sorted(batch_info.items()):
        channel_name = channel_data.channel_decoder.channel_name

        if len(channel_data.timestamps) < 2:
            continue  # not enough data to have lost anything

        lost_packets = 0
        lost_sections = 0
        for last, current, _i in channel_data.lost_packet_list:
            packets_lost = current - last - 1
            lost_packets += packets_lost
            lost_sections += 1

        lost_packets_file_boundaries = 0
        lost_sections_file_boundaries = 0
        for last, current, _i in channel_data.lost_packet_file_boundary_list:
            packets_lost = current - last - 1
            lost_packets_file_boundaries += packets_lost
            lost_sections_file_boundaries += 1

        if lost_packets != 0:
            m1 = "{}: Lost {} packet{} in {} place{}".format(
                channel_name,
                lost_packets, "s" if lost_packets != 1 else "",
                lost_sections, "s" if lost_sections != 1 else "")
            if lost_packets_file_boundaries != 0:
                m2 = "Lost {} packet{} in {} place{} between files".format(
                    lost_packets_file_boundaries,
                    "s" if lost_packets_file_boundaries != 1 else "",
                    lost_sections_file_boundaries,
                    "s" if lost_sections_file_boundaries != 1 else "")
                message = "\n".join([m1, m2, "Continue?"])
            else:
                message = "\n".join([m1, "Continue?"])
            logger.info(message)
            ret = QtWidgets.QMessageBox.warning(
                parent, "Warning", message,
                buttons=(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No),
                default=QtWidgets.QMessageBox.Yes)

            if ret == QtWidgets.QMessageBox.No:
                return False  # user aborted

            if once:
                return True

    return True


def sequence_data(batch_info, chunk=False, raw_units=False, parent=None):
    """
    returns a tuple, first element is a map keyed by channel index with a
    value of a list of tuples with (time, data, start_time, end_time) where
    time and data are numpy arrays of the same length. The second tuple element
    is a map keyed by channel index with a value of of map of unit strings in
    various formats, supporting keys of at least 'ascii', 'html', 'utf8'

    If chunk is False then a single NaNs will be inserted at any location where
    packets have been lost. This means that the tuple list will be length 1.

    If chunk is True, then each tuple will be broken up along lost packet
    boundaries.

    The first entry in the first time array will start at zero, regardless of
    its actual index value.

    The start_time value will be the UTC timestamp of the first decoded packet
    in the chunk. Likewise, end_time will be the UTC timestamp of the last
    decoded packet in the chunk
    """
    sequence_info = {}
    unit_info = {}
    for channel_index, channel_data in sorted(batch_info.items()):
        # create a time array with length of the original data array
        indexes = channel_data.indexes - channel_data.indexes[0]
        repeated_indexes = numpy.repeat(indexes.astype(numpy.double),
                                        channel_data.samples)
        fractional_section = (numpy.array(range(channel_data.samples),
                                          dtype=numpy.double) /
                              channel_data.samples)
        fractional_indexes = numpy.tile(fractional_section,
                                        channel_data.length)
        orig_time = ((repeated_indexes + fractional_indexes) /
                     channel_data.stream.rate)

        orig_data = channel_data.data

        # get the unit info now
        if raw_units:
            unit_formatter = asphodel.nativelib.create_unit_formatter(
                channel_data.channel.unit_type, 0.0, 0.0,
                channel_data.channel.resolution, use_metric=True)
        else:
            settings = QtCore.QSettings()
            unit_formatter = hyperborea.unit_preferences.create_unit_formatter(
                settings, channel_data.channel.unit_type,
                numpy.nanmin(orig_data), numpy.nanmax(orig_data),
                channel_data.channel.resolution)

        # save the units
        unit_info[channel_index] = {'ascii': unit_formatter.unit_ascii,
                                    'html': unit_formatter.unit_html,
                                    'utf8': unit_formatter.unit_utf8}

        # apply the scaling and offset
        orig_data = (orig_data * unit_formatter.conversion_scale +
                     unit_formatter.conversion_offset)

        if chunk:
            chunk_list = []

            last_data_index = 0
            last_timestamp_index = 0
            for i, (_x, _y, index) in enumerate(channel_data.lost_packet_list):
                data_index = index * channel_data.samples
                new_data = orig_data[last_data_index:data_index]
                new_time = orig_time[last_data_index:data_index]

                start = channel_data.timestamps[last_timestamp_index]
                end = channel_data.timestamps[index - 1]

                last_data_index = data_index
                last_timestamp_index = index

                chunk_list.append((new_time, new_data, start, end))

            # create a chunk for the last section
            new_data = orig_data[last_data_index:]
            new_time = orig_time[last_data_index:]

            start = channel_data.timestamps[last_timestamp_index]
            end = channel_data.timestamps[-1]

            chunk_list.append((new_time, new_data, start, end))

            sequence_info[channel_index] = chunk_list
        else:
            # create a new data array with space for the NaNs
            nan_count = len(channel_data.lost_packet_list)
            new_shape = (orig_data.shape[0] + nan_count, orig_data.shape[1])
            new_data = numpy.empty(new_shape)

            nan_slice = numpy.full((), numpy.nan, dtype=numpy.double)

            # create the new time array with spaces for NaNs
            new_time = numpy.empty((new_shape[0],))

            last = 0
            for i, (_x, _y, index) in enumerate(channel_data.lost_packet_list):
                index *= channel_data.samples
                new_data[last + i:index + i] = orig_data[last:index]
                new_data[index + i] = nan_slice
                new_time[last + i:index + i] = orig_time[last:index]
                new_time[index + i] = (orig_time[index - 1] +
                                       1 / (channel_data.stream.rate *
                                            channel_data.samples))
                last = index

            # copy in the last section
            new_data[last + nan_count:] = orig_data[last:]
            new_time[last + nan_count:] = orig_time[last:]

            # pull out start_time and end_time
            start_time = channel_data.timestamps[0]
            end_time = channel_data.timestamps[-1]

            sequence_info[channel_index] = [(new_time, new_data, start_time,
                                             end_time)]

    return sequence_info, unit_info


def choose_subchannel(channel_data, allow_all=False, parent=None):
    """
    returns a subchannel index selected by the user.

    If allow_all is set, it will present the user with a default "All
    Subchannels" option, which will return a value of -1 when selected.
    """

    options = list(channel_data.channel_decoder.subchannel_names)

    if len(options) == 1:
        return 0

    if allow_all:
        options.insert(0, "All Subchannels")

    value, ok = QtWidgets.QInputDialog.getItem(parent, "Select Subchannel",
                                               "Select Subchannel", options, 0,
                                               editable=False)
    if not ok:
        return  # user cancelled

    try:
        return channel_data.channel_decoder.subchannel_names.index(value)
    except ValueError:
        # must have been the "all subchannels" option
        return -1


def choose_subchannels(header, parent=None):
    """
    returns a sequence of tuples of the form (channel_index, subchannel_index)
    selected by the user.
    """

    info_list = []
    channel_indexes = set()
    for stream_id in header['streams_to_activate']:
        stream = header['streams'][stream_id]
        indexes = stream.channel_index_list[0:stream.channel_count]
        for index in indexes:
            channel_indexes.add(index)

        channel_list = [header['channels'][ch_id] for ch_id in indexes]
        info_list.append((stream_id, stream, channel_list))

    # create the device decoder
    decoder = asphodel.nativelib.create_device_decoder(
        info_list, header['stream_filler_bits'], header['stream_id_bits'])

    remaining_channels = set(channel_indexes)
    names = []  # list of (channel_id, channel_name, subchannel_names)
    try:
        for stream_decoder in decoder.decoders:
            for j, channel_decoder in enumerate(stream_decoder.decoders):
                channel_id = stream_decoder.stream_info.channel_index_list[j]
                if channel_id in remaining_channels:
                    remaining_channels.discard(channel_id)
                else:
                    continue

                names.append((channel_id, channel_decoder.channel_name,
                              channel_decoder.subchannel_names))
    except:
        message = "Could not create device decoder!"
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return None  # error

    dialog = SelectSubchannelsDialog(names, parent=parent)
    ret = dialog.exec_()
    if ret == 0:
        return  # user cancelled

    subchannels_list = dialog.get_subchannels_list()

    if len(subchannels_list) == 0:
        return None  # user didn't select any channels; treat like a cancel
    else:
        return subchannels_list


def get_contiguous_chunk(sequence, parent=None):
    if len(sequence) == 1:
        return sequence[0]  # nothing to do

    str_list = []
    str_indexes = {}

    max_points = 0
    max_index = 0

    for i, (time, _data, start_time, _end_time) in enumerate(sequence):
        points = len(time)

        if max_points < points:
            max_points = points
            max_index = i

        duration = time[-1] - time[0]  # close enough
        start_dt = datetime.datetime.fromtimestamp(start_time,
                                                   datetime.timezone.utc)
        start_str = start_dt.strftime("%Y-%m-%d %H:%M:%S (UTC)")
        s = "{} points ({} s), starting {}".format(points, duration, start_str)
        if s not in str_list:
            # NOTE: this prevents selection of duplicate chunks small enough
            # to have the same start time, but in reality, this isn't a
            # realistic use case
            str_list.append(s)
            str_indexes[s] = i

    value, ok = QtWidgets.QInputDialog.getItem(
        parent, "Select Contiguous Section", "Select Section", str_list,
        max_index, editable=False)
    if not ok:
        return  # user cancelled

    return sequence[str_indexes[value]]


def get_psd_options(chunks_list, sampling_rate, parent=None):
    dialog = PSDOptionsDialog(chunks_list, sampling_rate)

    ret = dialog.exec_()
    if ret == 0:
        return  # user cancelled

    return dialog.get_options()


def get_multiple_psd_options(sections, parent=None):
    dialog = MultiplePSDOptionsDialog(sections)

    ret = dialog.exec_()
    if ret == 0:
        return  # user cancelled

    return dialog.get_options()



#!/usr/bin/env python3
import datetime
import distutils.core
import glob
import os.path
import shutil
import subprocess
import sys

import cx_Freeze
import matplotlib.backends
import PySide2
import requests


is_64bit = sys.maxsize > (2 ** 32)

cwd = os.path.abspath(os.path.dirname(__file__))
version = subprocess.check_output([sys.executable, "setup.py", "--version"],
                                  cwd=cwd, universal_newlines=True).strip()

if sys.platform == "win32":
    if is_64bit:
        build_key = "win64"
    else:
        build_key = "win32"
else:
    build_key = "unknown"


# use git to find the branch name and commit hash
try:
    branch = subprocess.check_output(
        ['git', 'rev-parse', '--abbrev-ref', 'HEAD'], universal_newlines=True,
        cwd=cwd).strip()
    commit = subprocess.check_output(
        ['git', 'rev-parse', 'HEAD'], universal_newlines=True, cwd=cwd).strip()
except:
    # git probably isn't installed
    branch = ""
    commit = ""
build_date = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
with open(os.path.join(cwd, "build_info.txt"), "w") as f:
    f.write("{}\n{}\n{}\n{}\n".format(branch, commit, build_key, build_date))

# GUI applications require a different base on Windows (the default is for a
# console application).
base = None
if sys.platform == "win32":
    base = "Win32GUI"

if is_64bit:
    setup_include = ('install_extras/mondo_64bit.iss', 'mondo.iss')
else:
    setup_include = ('install_extras/mondo_32bit.iss', 'mondo.iss')

includefiles = [("build_info.txt", "build_info.txt"),
                ("mondo/mondo.ico", "mondo.ico"),
                (requests.certs.where(), 'cacert.pem'),
                setup_include,
                ('LICENSE.txt', 'LICENSE.txt')]

pyside2_prefix = os.path.abspath(os.path.dirname(PySide2.__file__))
pyside2_includefiles = [
    (pyside2_prefix + "/plugins/platforms", "platforms"),
    (pyside2_prefix + "/plugins/styles", "styles"),
    (pyside2_prefix + "/plugins/imageformats/qico.dll",
     "imageformats/qico.dll")]
includefiles.extend(pyside2_includefiles)


includes = ["atexit",  # for QT, not autodetected by cx_freeze
            'numpy.core._methods', 'numpy.lib.format',
            'idna', 'idna.idnadata',
            'scipy._distributor_init', 'scipy.sparse.csgraph._validation',
            'scipy.fftpack._fftpack', 'scipy.ndimage._ni_support',
            'matplotlib.backends',
            'matplotlib.backends.backend_qt4agg']
excludes = ["pysideuic",  # pulled in by pyqtgraph, not needed
            "scipy.spatial.cKDTree",  # work around for case sensitive issue
            "tkinter",
            'ipykernel', "IPython", "jupyter",
            'PySide',
            ]
packages = []


backend_path = os.path.dirname(matplotlib.backends.__file__)
backend_files = glob.glob(os.path.join(backend_path, 'backend_*'))
for f in backend_files:
    module = os.path.splitext(os.path.basename(f))[0]
    if not module.startswith("backend_qt"):
        if not module.startswith("backend_agg"):
            full_module = "matplotlib.backends." + module
            excludes.append(full_module)


class BdistInnoCommand(distutils.core.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def _find_iscc(self):
        paths = []
        program_files_keys = ['PROGRAMW6432', 'PROGRAMFILES',
                              'PROGRAMFILES(X86)']
        program_files_dirs = []
        for key in program_files_keys:
            try:
                path = os.environ[key]
                if path:
                    program_files_dirs.append(path)
            except KeyError:
                pass
        for program_files in program_files_dirs:
            paths.append(os.path.join(program_files, "Inno Setup 6"))
        paths.append(os.environ['PATH'])  # not extend; it's a string
        path = os.pathsep.join(paths)
        return shutil.which("iscc", path=path)

    def run(self):
        build = self.get_finalized_command("build")
        build.run()
        iscc = self._find_iscc()
        if not iscc:
            raise Exception("Could not find ISCC.exe!")

        iss_files = glob.glob(os.path.join(build.build_exe, "*.iss"))
        if not iss_files:
            raise Exception("No iss file in build directory!")
        elif len(iss_files) > 1:
            raise Exception("Too many iss files in build directory!")

        iss_file = iss_files[0]

        dist_dir = "dist"

        subprocess.check_call([iscc, "/DVERSION=" + version, "/O" + dist_dir,
                               iss_file], cwd=cwd)


if sys.platform == "win32":
    cmdclass = {'bdist_inno': BdistInnoCommand}
else:
    cmdclass = {}


cx_Freeze.setup(
    name="mondo",
    version=version,
    author="Suprock Technologies, LLC",
    author_email="inquiries@suprocktech.com",
    description="Application for analyzing recorded Asphodel sensor data",
    options={"build_exe": {"includes": includes,
                           "excludes": excludes,
                           "packages": packages,
                           "include_files": includefiles,
                           "zip_include_packages": ['*'],
                           'zip_exclude_packages': ['numpy', 'scipy',
                                                    'asphodel'],
                           "include_msvcr": True,
                           "replace_paths": [("*", "")]}},
    executables=[cx_Freeze.Executable(script="cx_shim_run.py", base=base,
                                      targetName="mondo.exe",
                                      icon="mondo/mondo.ico")],
    cmdclass=cmdclass,
)



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging

from PySide2 import QtWidgets

from . import util
from .. import export_script

logger = logging.getLogger(__name__)


def spectrogram_analysis(parent=None):
    ret = util.load_batch(parent=parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    channel_index = util.choose_channel(header, parent=parent)
    if channel_index is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, [channel_index],
                                   parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    subchannel_index = util.choose_subchannel(batch_info[channel_index],
                                              allow_all=False)
    if subchannel_index is None:
        return  # user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, parent=parent)
    if not ret:
        return  # user cancelled

    sequence_info, _unit_info = util.sequence_data(
        batch_info, chunk=False, raw_units=True, parent=parent)
    chunked_sequence_info, _unit_info = util.sequence_data(
        batch_info, chunk=True, raw_units=True, parent=parent)
    chunk_list = chunked_sequence_info[channel_index]
    logger.debug("%s", chunk_list)

    sequence = sequence_info[channel_index]

    time, data, _start, _end = sequence[0]

    if len(time) < 256:
        # not enough data to make a decent PSD
        message = "Not enough data points! Only {}!".format(len(time))
        logger.error(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return

    sampling_rate = 1.0 / (time[1] - time[0])

    psd_options = util.get_psd_options([chunk_list], sampling_rate,
                                       parent=parent)
    if psd_options is None:
        return  # user cancelled

    data = data[:, subchannel_index]

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    try:
        ax.specgram(data, **psd_options)
    except MemoryError:
        message = "Out of memory!"
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return

    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Frequency (Hz)")

    ax.autoscale(tight=True)

    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    ax.get_yaxis().get_major_formatter().set_useOffset(False)

    fig.show()
